from __future__ import annotations

import ast
import asyncio
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    List,
    Optional,
    Type,
    Union,
    cast,
)

from ....utils.async_itertools import async_next
from ....utils.async_tools import threaded
from ....utils.logging import LoggingDescriptor
from ....utils.uri import Uri
from ...common.decorators import language_id
from ...common.lsp_types import Location, LocationLink, Position
from ...common.text_document import TextDocument
from ..utils.ast import (
    HasTokens,
    Token,
    get_nodes_at_position,
    get_tokens_at_position,
    range_from_token,
)

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol

from .model_helper import ModelHelperMixin
from .protocol_part import RobotLanguageServerProtocolPart


class CollectType(Enum):
    DEFINITION = 1
    IMPLEMENTATION = 2


_CollectMethod = Callable[
    [ast.AST, List[ast.AST], TextDocument, Position, CollectType],
    Awaitable[Union[Location, List[Location], List[LocationLink], None]],
]


class RobotGotoProtocolPart(RobotLanguageServerProtocolPart, ModelHelperMixin):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        parent.definition.collect.add(self.collect_definition)
        parent.implementation.collect.add(self.collect_implementation)

    def _find_method(self, cls: Type[Any]) -> Optional[_CollectMethod]:
        if cls is ast.AST:
            return None
        method_name = "definition_" + cls.__name__
        if hasattr(self, method_name):
            method = getattr(self, method_name)
            if callable(method):
                return cast(_CollectMethod, method)
        for base in cls.__bases__:
            method = self._find_method(base)
            if method:
                return cast(_CollectMethod, method)
        return None

    @language_id("robotframework")
    @threaded()
    @_logger.call
    async def collect_definition(
        self, sender: Any, document: TextDocument, position: Position
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        return await self.collect(document, position, CollectType.DEFINITION)

    @language_id("robotframework")
    @threaded()
    @_logger.call
    async def collect_implementation(
        self, sender: Any, document: TextDocument, position: Position
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        return await self.collect(document, position, CollectType.IMPLEMENTATION)

    @_logger.call
    async def collect(
        self, document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        result_nodes = await get_nodes_at_position(await self.parent.documents_cache.get_model(document), position)

        if not result_nodes:
            return None

        result_node = result_nodes[-1]

        if result_node is None:
            return None

        result = await self._definition_default(result_nodes, document, position, collect_type)
        if result:
            return result

        method = self._find_method(type(result_node))
        if method is not None:
            result = await method(result_node, result_nodes, document, position, collect_type)
            if result is not None:
                return result

        return None

    async def _definition_default(
        self, nodes: List[ast.AST], document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        if not nodes:
            return None

        node = nodes[-1]

        if not isinstance(node, HasTokens):
            return None

        tokens = get_tokens_at_position(node, position)

        for token in tokens:
            token_and_var = await async_next(
                (
                    (var_token, var)
                    async for var_token, var in self.iter_variables_from_token(
                        token, namespace, nodes, position, skip_commandline_variables=True
                    )
                    if position in range_from_token(var_token)
                ),
                None,
            )

            if token_and_var is not None:
                var_token, variable = token_and_var
                if variable.source:
                    return [
                        LocationLink(
                            origin_selection_range=range_from_token(var_token),
                            target_uri=str(Uri.from_path(variable.source)),
                            target_range=variable.range(),
                            target_selection_range=range_from_token(variable.name_token)
                            if variable.name_token
                            else variable.range(),
                        )
                    ]
        return None

    async def definition_IfHeader(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import IfHeader

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        header = cast(IfHeader, node)

        expression_token = header.get_token(RobotToken.ARGUMENT)
        if expression_token is not None and position in range_from_token(expression_token):
            token_and_var = await async_next(
                (
                    (var_token, var)
                    async for var_token, var in self.iter_expression_variables_from_token(
                        expression_token, namespace, nodes, position, skip_commandline_variables=True
                    )
                    if position in range_from_token(var_token)
                ),
                None,
            )
            if token_and_var is not None:
                var_token, variable = token_and_var

                if variable.source:
                    return [
                        LocationLink(
                            origin_selection_range=range_from_token(var_token),
                            target_uri=str(Uri.from_path(variable.source)),
                            target_range=variable.range(),
                            target_selection_range=range_from_token(variable.name_token)
                            if variable.name_token
                            else variable.range(),
                        )
                    ]

        return None

    async def definition_WhileHeader(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import WhileHeader

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        header = cast(WhileHeader, node)

        expression_token = header.get_token(RobotToken.ARGUMENT)
        if expression_token is not None and position in range_from_token(expression_token):
            token_and_var = await async_next(
                (
                    (var_token, var)
                    async for var_token, var in self.iter_expression_variables_from_token(
                        expression_token, namespace, nodes, position, skip_commandline_variables=True
                    )
                    if position in range_from_token(var_token)
                ),
                None,
            )
            if token_and_var is not None:
                var_token, variable = token_and_var

                if variable.source:
                    return [
                        LocationLink(
                            origin_selection_range=range_from_token(var_token),
                            target_uri=str(Uri.from_path(variable.source)),
                            target_range=variable.range(),
                            target_selection_range=range_from_token(variable.name_token)
                            if variable.name_token
                            else variable.range(),
                        )
                    ]
        return None

    async def definition_KeywordName(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import KeywordName

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        kw_node = cast(KeywordName, node)
        name_token = cast(RobotToken, kw_node.get_token(RobotToken.KEYWORD_NAME))

        if not name_token:
            return None

        result = await namespace.find_keyword(name_token.value)

        if result is not None and not result.is_error_handler and result.source:
            token_range = range_from_token(name_token)
            return [
                LocationLink(
                    origin_selection_range=token_range,
                    target_uri=str(Uri.from_path(result.source)),
                    target_range=token_range,
                    target_selection_range=token_range,
                )
            ]

        return None

    async def definition_KeywordCall(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import KeywordCall

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        kw_node = cast(KeywordCall, node)
        result = await self.get_keyworddoc_and_token_from_position(
            kw_node.keyword,
            cast(Token, kw_node.get_token(RobotToken.KEYWORD)),
            [cast(Token, t) for t in kw_node.get_tokens(RobotToken.ARGUMENT)],
            namespace,
            position,
        )

        if result is not None:
            keyword_doc, keyword_token = result

            keyword_token = self.strip_bdd_prefix(keyword_token)

            lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

            kw_range = range_from_token(keyword_token)

            if lib_entry and kw_namespace:
                r = range_from_token(keyword_token)
                r.end.character = r.start.character + len(kw_namespace)
                kw_range.start.character = r.end.character + 1
                if position in r:
                    if collect_type == CollectType.DEFINITION and lib_entry.import_source:
                        return [
                            LocationLink(
                                origin_selection_range=r,
                                target_uri=str(Uri.from_path(lib_entry.import_source)),
                                target_range=lib_entry.import_range,
                                target_selection_range=lib_entry.import_range,
                            )
                        ]
                    if lib_entry.library_doc and lib_entry.library_doc.source_or_origin:
                        return [
                            LocationLink(
                                origin_selection_range=r,
                                target_uri=str(Uri.from_path(lib_entry.library_doc.source_or_origin)),
                                target_range=lib_entry.import_range,
                                target_selection_range=lib_entry.import_range,
                            )
                        ]
                    else:
                        return None

            if position in kw_range and keyword_doc is not None and keyword_doc.source:
                return [
                    LocationLink(
                        origin_selection_range=kw_range,
                        target_uri=str(Uri.from_path(keyword_doc.source)),
                        target_range=keyword_doc.range,
                        target_selection_range=keyword_doc.range,
                    )
                ]

        return None

    async def definition_Fixture(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Fixture

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        fixture_node = cast(Fixture, node)

        name_token = cast(Token, fixture_node.get_token(RobotToken.NAME))
        if name_token is None or name_token.value is None or name_token.value in ("", "NONE"):
            return None

        result = await self.get_keyworddoc_and_token_from_position(
            fixture_node.name,
            name_token,
            [cast(Token, t) for t in fixture_node.get_tokens(RobotToken.ARGUMENT)],
            namespace,
            position,
        )

        if result is not None:
            keyword_doc, keyword_token = result

            keyword_token = self.strip_bdd_prefix(keyword_token)

            lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

            kw_range = range_from_token(keyword_token)

            if lib_entry and kw_namespace:
                r = range_from_token(keyword_token)
                r.end.character = r.start.character + len(kw_namespace)
                kw_range.start.character = r.end.character + 1
                if position in r:
                    if collect_type == CollectType.DEFINITION and lib_entry.import_source:
                        return [
                            LocationLink(
                                origin_selection_range=r,
                                target_uri=str(Uri.from_path(lib_entry.import_source)),
                                target_range=lib_entry.import_range,
                                target_selection_range=lib_entry.import_range,
                            )
                        ]
                    if lib_entry.library_doc and lib_entry.library_doc.source_or_origin:
                        return [
                            LocationLink(
                                origin_selection_range=r,
                                target_uri=str(Uri.from_path(lib_entry.library_doc.source_or_origin)),
                                target_range=lib_entry.import_range,
                                target_selection_range=lib_entry.import_range,
                            )
                        ]
                    else:
                        return None

            if position in kw_range and keyword_doc is not None and keyword_doc.source:
                return [
                    LocationLink(
                        origin_selection_range=kw_range,
                        target_uri=str(Uri.from_path(keyword_doc.source)),
                        target_range=keyword_doc.range,
                        target_selection_range=keyword_doc.range,
                    )
                ]

        return None

    async def _definition_Template_or_TestTemplate(  # noqa: N802
        self,
        template_node: ast.AST,
        nodes: List[ast.AST],
        document: TextDocument,
        position: Position,
        collect_type: CollectType,
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Template, TestTemplate

        template_node = cast(Union[Template, TestTemplate], template_node)
        if template_node.value:

            keyword_token = cast(RobotToken, template_node.get_token(RobotToken.NAME))
            if keyword_token is None:
                return None

            keyword_token = self.strip_bdd_prefix(keyword_token)

            if position.is_in_range(range_from_token(keyword_token)):
                namespace = await self.parent.documents_cache.get_namespace(document)
                if namespace is None:
                    return None

                keyword_doc = await namespace.find_keyword(template_node.value)
                if keyword_doc is not None:

                    lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

                    kw_range = range_from_token(keyword_token)

                    if lib_entry and kw_namespace:
                        r = range_from_token(keyword_token)
                        r.end.character = r.start.character + len(kw_namespace)
                        kw_range.start.character = r.end.character + 1
                        if position in r:
                            if collect_type == CollectType.DEFINITION and lib_entry.import_source:
                                return [
                                    LocationLink(
                                        origin_selection_range=r,
                                        target_uri=str(Uri.from_path(lib_entry.import_source)),
                                        target_range=lib_entry.import_range,
                                        target_selection_range=lib_entry.import_range,
                                    )
                                ]
                            if lib_entry.library_doc and lib_entry.library_doc.source_or_origin:
                                return [
                                    LocationLink(
                                        origin_selection_range=r,
                                        target_uri=str(Uri.from_path(lib_entry.library_doc.source_or_origin)),
                                        target_range=lib_entry.import_range,
                                        target_selection_range=lib_entry.import_range,
                                    )
                                ]
                            else:
                                return None

                    if keyword_doc.source and not keyword_doc.is_error_handler:
                        return [
                            LocationLink(
                                origin_selection_range=kw_range,
                                target_uri=str(Uri.from_path(keyword_doc.source)),
                                target_range=keyword_doc.range,
                                target_selection_range=keyword_doc.range,
                            )
                        ]
        return None

    async def definition_TestTemplate(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        return await self._definition_Template_or_TestTemplate(node, nodes, document, position, collect_type)

    async def definition_Template(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        return await self._definition_Template_or_TestTemplate(node, nodes, document, position, collect_type)

    async def definition_LibraryImport(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import LibraryImport

        library_node = cast(LibraryImport, node)
        if library_node.name:

            name_token = cast(RobotToken, library_node.get_token(RobotToken.NAME))
            if name_token is None:
                return None

            if position.is_in_range(range_from_token(name_token)):
                namespace = await self.parent.documents_cache.get_namespace(document)
                if namespace is None:
                    return None

                try:
                    libdoc = await namespace.get_imported_library_libdoc(
                        library_node.name, library_node.args, library_node.alias
                    )

                    if libdoc is None or libdoc.errors:
                        libdoc = await namespace.imports_manager.get_libdoc_for_library_import(
                            str(library_node.name), (), str(document.uri.to_path().parent)
                        )

                    if libdoc is None:
                        return None

                    python_source = libdoc.source_or_origin
                    if python_source is not None:
                        return [
                            LocationLink(
                                origin_selection_range=range_from_token(name_token),
                                target_uri=str(Uri.from_path(python_source)),
                                target_range=libdoc.range,
                                target_selection_range=libdoc.range,
                            )
                        ]
                except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
                    raise
                except BaseException:
                    pass
        return None

    async def definition_ResourceImport(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import ResourceImport

        resource_node = cast(ResourceImport, node)
        if resource_node.name:

            name_token = cast(RobotToken, resource_node.get_token(RobotToken.NAME))
            if name_token is None:
                return None

            if position.is_in_range(range_from_token(name_token)):
                namespace = await self.parent.documents_cache.get_namespace(document)
                if namespace is None:
                    return None

                try:
                    libdoc = await namespace.get_imported_resource_libdoc(resource_node.name)

                    if libdoc is None or libdoc.errors:
                        libdoc = await namespace.imports_manager.get_libdoc_for_resource_import(
                            str(resource_node.name), str(document.uri.to_path().parent)
                        )

                    if libdoc is None:
                        return None

                    python_source = libdoc.source_or_origin
                    if python_source is not None:
                        return [
                            LocationLink(
                                origin_selection_range=range_from_token(name_token),
                                target_uri=str(Uri.from_path(python_source)),
                                target_range=libdoc.range,
                                target_selection_range=libdoc.range,
                            )
                        ]
                except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
                    raise
                except BaseException:
                    pass
        return None

    async def definition_VariablesImport(  # noqa: N802
        self, node: ast.AST, nodes: List[ast.AST], document: TextDocument, position: Position, collect_type: CollectType
    ) -> Union[Location, List[Location], List[LocationLink], None]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import VariablesImport

        variables_node = cast(VariablesImport, node)
        if variables_node.name:

            name_token = cast(RobotToken, variables_node.get_token(RobotToken.NAME))
            if name_token is None:
                return None

            if position.is_in_range(range_from_token(name_token)):
                namespace = await self.parent.documents_cache.get_namespace(document)
                if namespace is None:
                    return None

                try:
                    libdoc = await namespace.get_imported_variables_libdoc(variables_node.name, variables_node.args)

                    if libdoc is None or libdoc.errors:
                        libdoc = await namespace.imports_manager.get_libdoc_for_variables_import(
                            str(variables_node.name), (), str(document.uri.to_path().parent)
                        )

                    if libdoc is None:
                        return None

                    python_source = libdoc.source_or_origin
                    if python_source is not None:
                        return [
                            LocationLink(
                                origin_selection_range=range_from_token(name_token),
                                target_uri=str(Uri.from_path(python_source)),
                                target_range=libdoc.range,
                                target_selection_range=libdoc.range,
                            )
                        ]
                except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
                    raise
                except BaseException:
                    pass
        return None



import os
from typing import Dict, Mapping, Optional

from robot.variables.finders import (
    EmptyFinder,
    NumberFinder,
    StoredFinder,
    VariableFinder,
    variable_not_found,
)
from robot.variables.replacer import VariableReplacer
from robot.variables.store import VariableStore


class EnvironmentFinder:
    identifiers = "%"

    def __init__(self, env: Optional[Mapping[str, str]] = None) -> None:
        self.env = env if env is not None else dict(os.environ)

    def find(self, name: str) -> Optional[str]:  # type: ignore
        var_name, has_default, default_value = name[2:-1].partition("=")

        value = self.env.get(var_name, None)
        if value is not None:
            return value
        if has_default:  # in case if '' is desired default value
            return default_value
        variable_not_found(name, self._get_candidates(), "Environment variable '%s' not found." % name)

    def _get_candidates(self) -> Mapping[str, str]:
        candidates: Dict[str, str] = dict()
        candidates.update(self.env)
        return candidates


class RobotCodeVariableFinder(VariableFinder):
    def __init__(self, variable_store: VariableStore, env: Optional[Mapping[str, str]] = None) -> None:
        super().__init__(variable_store)
        self._finders = (
            StoredFinder(variable_store),
            NumberFinder(),
            EmptyFinder(),
            EnvironmentFinder(env),
        )


class RobotCodeVariableReplacer(VariableReplacer):
    def __init__(self, variable_store: VariableStore, env: Optional[Mapping[str, str]] = None) -> None:
        super().__init__(variable_store)
        self._finder = RobotCodeVariableFinder(variable_store, env)



import socket
from contextlib import closing
from typing import cast

from .logging import LoggingDescriptor

_logger = LoggingDescriptor(name=__name__)


def find_free_port() -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("127.0.0.1", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return cast(int, s.getsockname()[1])


def check_free_port(port: int) -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        try:
            s.bind(("127.0.0.1", port))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return cast(int, s.getsockname()[1])
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException:
            _logger.warning(f"Port {port} is not free. Try to find a free port.")
            return find_free_port()



from __future__ import annotations

import ast
from typing import TYPE_CHECKING, Any, List, Optional, cast

from ....utils.async_tools import threaded
from ....utils.logging import LoggingDescriptor
from ...common.decorators import language_id
from ...common.lsp_types import CodeLens, Command
from ...common.text_document import TextDocument
from ..utils.ast import range_from_token

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol

from .protocol_part import RobotLanguageServerProtocolPart


class RobotCodeLensProtocolPart(RobotLanguageServerProtocolPart):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        # parent.code_lens.collect.add(self.collect)

    @language_id("robotframework")
    @threaded()
    async def collect(self, sender: Any, document: TextDocument) -> Optional[List[CodeLens]]:

        from ..utils.async_ast import AsyncVisitor

        class Visitor(AsyncVisitor):
            def __init__(self, parent: RobotCodeLensProtocolPart) -> None:
                super().__init__()
                self.parent = parent

                self.result: List[CodeLens] = []

            async def visit(self, node: ast.AST) -> None:
                await super().visit(node)

            @classmethod
            async def find_from(cls, model: ast.AST, parent: RobotCodeLensProtocolPart) -> Optional[List[CodeLens]]:

                finder = cls(parent)

                await finder.visit(model)

                return finder.result if finder.result else None

            async def visit_Section(self, node: ast.AST) -> None:  # noqa: N802
                from robot.parsing.model.blocks import KeywordSection

                if isinstance(node, KeywordSection):
                    await self.generic_visit(node)

            async def visit_Keyword(self, node: ast.AST) -> None:  # noqa: N802
                from robot.parsing.lexer.tokens import Token as RobotToken
                from robot.parsing.model.blocks import Keyword

                keyword = cast(Keyword, node)

                if keyword.header:
                    name_token = keyword.header.get_token(RobotToken.KEYWORD_NAME)
                    if name_token is None:
                        return

                    r = range_from_token(name_token)
                    self.result.append(
                        CodeLens(
                            r,
                            Command(
                                "references",
                                "robotcode.action.findReferences",
                                [str(document.uri), {"lineNumber": r.start.line, "column": r.start.character}],
                            ),
                        )
                    )

        return await Visitor.find_from(await self.parent.documents_cache.get_model(document), self)



from __future__ import annotations

import asyncio
import os
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

from ...jsonrpc2.protocol import rpc_method
from ...jsonrpc2.server import JsonRPCServer, JsonRpcServerMode, TcpParams
from ...utils.logging import LoggingDescriptor
from ..client import DAPClient, DAPClientError
from ..dap_types import (
    Capabilities,
    ConfigurationDoneArguments,
    ConfigurationDoneRequest,
    DisconnectArguments,
    DisconnectRequest,
    ExceptionBreakpointsFilter,
    InitializeRequestArguments,
    LaunchRequestArguments,
    OutputCategory,
    OutputEvent,
    OutputEventBody,
    Request,
    RunInTerminalKind,
    RunInTerminalRequest,
    RunInTerminalRequestArguments,
    RunInTerminalResponseBody,
    TerminateArguments,
    TerminatedEvent,
    TerminateRequest,
)
from ..protocol import DebugAdapterProtocol


class OutputProtocol(asyncio.SubprocessProtocol):
    def __init__(self, parent: LauncherDebugAdapterProtocol) -> None:
        super().__init__()
        self.parent = parent

    def pipe_data_received(self, fd: Any, data: bytes) -> None:
        category = None

        if fd == 1:
            category = OutputCategory.STDOUT
        elif fd == 2:
            category = OutputCategory.STDERR

        self.parent.send_event(OutputEvent(body=OutputEventBody(output=data.decode(), category=category)))


class LauncherDebugAdapterProtocol(DebugAdapterProtocol):
    _logger = LoggingDescriptor()

    def __init__(self) -> None:
        super().__init__()
        self._client: Optional[DAPClient] = None
        self._process: Optional[asyncio.subprocess.Process] = None

    @property
    def client(self) -> DAPClient:
        if self._client is None:
            raise DAPClientError("Client not defined.")

        return self._client

    @client.setter
    def client(self, value: DAPClient) -> None:
        self._client = value

    @property
    def connected(self) -> bool:
        return self._client is not None and self._client.connected

    @rpc_method(name="initialize", param_type=InitializeRequestArguments)
    async def _initialize(self, arguments: InitializeRequestArguments, *args: Any, **kwargs: Any) -> Capabilities:
        self._initialized = True

        return Capabilities(
            supports_configuration_done_request=True,
            supports_conditional_breakpoints=True,
            supports_hit_conditional_breakpoints=True,
            support_terminate_debuggee=True,
            support_suspend_debuggee=True,
            supports_evaluate_for_hovers=True,
            supports_terminate_request=True,
            supports_log_points=True,
            supports_set_expression=True,
            supports_set_variable=True,
            supports_value_formatting_options=True,
            exception_breakpoint_filters=[
                ExceptionBreakpointsFilter(
                    filter="failed_keyword",
                    label="Failed Keywords",
                    description="Breaks on failed keywords",
                    default=True,
                    # supports_condition=True,
                    # condition_description="expression",
                ),
                ExceptionBreakpointsFilter(
                    filter="failed_test",
                    label="Failed Test",
                    description="Breaks on failed tests",
                    default=False,
                    # supports_condition=True,
                    # condition_description="expression",
                ),
                ExceptionBreakpointsFilter(
                    filter="failed_suite",
                    label="Failed Suite",
                    description="Breaks on failed suite",
                    default=False,
                    # supports_condition=True,
                    # condition_description="expression",
                ),
            ],
            supports_exception_options=True,
            supports_exception_filter_options=True,
        )

    @rpc_method(name="launch", param_type=LaunchRequestArguments)
    async def _launch(
        self,
        request: str,
        python: str,
        cwd: str = ".",
        target: Optional[str] = None,
        args: Optional[List[str]] = None,
        env: Optional[Dict[str, Optional[Any]]] = None,
        console: Optional[Literal["internalConsole", "integratedTerminal", "externalTerminal"]] = "integratedTerminal",
        name: Optional[str] = None,
        no_debug: Optional[bool] = None,
        robotPythonPath: Optional[List[str]] = None,  # noqa: N803
        launcherArgs: Optional[List[str]] = None,  # noqa: N803
        launcherTimeout: Optional[int] = None,  # noqa: N803
        attachPython: Optional[bool] = False,  # noqa: N803
        variables: Optional[Dict[str, Any]] = None,
        outputDir: Optional[str] = None,
        outputMessages: Optional[bool] = False,
        outputLog: Optional[bool] = False,
        groupOutput: Optional[bool] = False,
        stopOnEntry: Optional[bool] = False,  # noqa: N803
        arguments: Optional[LaunchRequestArguments] = None,
        dryRun: Optional[bool] = None,
        *_args: Any,
        **_kwargs: Any,
    ) -> None:
        from ...utils.net import find_free_port

        connect_timeout = launcherTimeout or 5

        port = find_free_port()

        launcher = Path(Path(__file__).parent.parent)

        run_args = [python, "-u", str(launcher)]

        run_args += ["-p", str(port)]
        run_args += ["-w", "-t", str(connect_timeout)]

        if no_debug:
            run_args += ["-n"]

        if attachPython and not no_debug:
            run_args += ["-d", "-dp", str(find_free_port()), "-dw"]

        if outputMessages:
            run_args += ["-om"]

        if outputLog:
            run_args += ["-ol"]

        if groupOutput:
            run_args += ["-og"]

        if stopOnEntry:
            run_args += ["-soe"]

        run_args += launcherArgs or []

        run_args += ["--"]

        if dryRun:
            run_args += ["--dryrun"]

        if outputDir:
            run_args += ["-d", outputDir]

        if robotPythonPath:
            for e in robotPythonPath:
                run_args += ["-P", e]

        if variables:
            for k, v in variables.items():
                run_args += ["-v", f"{k}:{v}"]

        run_args += args or []

        if target:
            run_args.append(target)

        env = {k: ("" if v is None else str(v)) for k, v in env.items()} if env else {}

        if console in ["integratedTerminal", "externalTerminal"]:
            await self.send_request_async(
                RunInTerminalRequest(
                    arguments=RunInTerminalRequestArguments(
                        cwd=cwd,
                        args=run_args,
                        env=env,
                        kind=RunInTerminalKind.INTEGRATED
                        if console == "integratedTerminal"
                        else RunInTerminalKind.EXTERNAL
                        if console == "externalTerminal"
                        else None,
                        title=name,
                    )
                ),
                return_type=RunInTerminalResponseBody,
            )
        elif console is None or console in ["internalConsole"]:
            run_env: Dict[str, Optional[str]] = dict(os.environ)
            run_env.update(env)

            await asyncio.get_event_loop().subprocess_exec(
                lambda: OutputProtocol(self),
                *run_args,
                cwd=cwd,
                env=run_env,
            )

        else:
            raise Exception(f'Unknown console type "{console}".')

        self.client = DAPClient(self, TcpParams(None, port))
        try:
            await self.client.connect(connect_timeout)
        except asyncio.TimeoutError:
            raise asyncio.TimeoutError("Can't connect to debug launcher.")

    @rpc_method(name="configurationDone", param_type=ConfigurationDoneArguments)
    async def _configuration_done(
        self, arguments: Optional[ConfigurationDoneArguments] = None, *args: Any, **kwargs: Any
    ) -> None:
        await self.client.protocol.send_request_async(ConfigurationDoneRequest(arguments=arguments))

    @rpc_method(name="disconnect", param_type=DisconnectArguments)
    async def _disconnect(self, arguments: Optional[DisconnectArguments] = None, *args: Any, **kwargs: Any) -> None:
        if self.connected:
            if not self.client.protocol.terminated:
                await self.client.protocol.send_request_async(DisconnectRequest(arguments=arguments))
        else:
            await self.send_event_async(TerminatedEvent())

    @_logger.call
    @rpc_method(name="terminate", param_type=TerminateArguments)
    async def _terminate(self, arguments: Optional[TerminateArguments] = None, *args: Any, **kwargs: Any) -> None:
        if self.client.connected:
            return await self.client.protocol.send_request_async(TerminateRequest(arguments=arguments))
        else:
            await self.send_event_async(TerminatedEvent())

    async def handle_unknown_command(self, message: Request) -> Any:
        if self.connected:
            self._logger.info("Forward request to client...")

            return await self.client.protocol.send_request_async(message)

        return await super().handle_unknown_command(message)


TCP_DEFAULT_PORT = 6611


class LauncherServer(JsonRPCServer[LauncherDebugAdapterProtocol]):
    def __init__(
        self,
        mode: JsonRpcServerMode = JsonRpcServerMode.STDIO,
        tcp_params: TcpParams = TcpParams(None, TCP_DEFAULT_PORT),
    ):
        super().__init__(
            mode=mode,
            tcp_params=tcp_params,
        )

    def create_protocol(self) -> LauncherDebugAdapterProtocol:
        return LauncherDebugAdapterProtocol()



from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

from ....utils.logging import LoggingDescriptor
from ....utils.uri import Uri
from ...common.lsp_types import DocumentUri
from ...common.text_document import TextDocument

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol

from .protocol_part import RobotLanguageServerProtocolPart


class RobotWorkspaceProtocolPart(RobotLanguageServerProtocolPart):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

    @_logger.call
    async def get_or_open_document(
        self, path: Union[str, os.PathLike[Any]], language_id: str, version: Optional[int] = None
    ) -> TextDocument:
        from robot.utils import FileReader

        uri = Uri.from_path(path).normalized()

        result = await self.parent.documents.get(uri)
        if result is not None:
            return result

        with FileReader(Path(path)) as reader:
            text = str(reader.read())

        return await self.parent.documents.append_document(
            document_uri=DocumentUri(uri), language_id=language_id, text=text, version=version
        )



from __future__ import annotations

import gc
import re
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Union

from ....jsonrpc2.protocol import JsonRPCException, rpc_method
from ....utils.async_tools import Lock, async_tasking_event
from ....utils.logging import LoggingDescriptor
from ....utils.uri import Uri
from ..decorators import language_id_filter
from ..lsp_types import (
    DidChangeTextDocumentParams,
    DidCloseTextDocumentParams,
    DidOpenTextDocumentParams,
    DidSaveTextDocumentParams,
    DocumentUri,
    FileEvent,
    TextDocumentContentChangeEvent,
    TextDocumentContentRangeChangeEvent,
    TextDocumentContentTextChangeEvent,
    TextDocumentIdentifier,
    TextDocumentItem,
    TextDocumentSaveReason,
    TextDocumentSyncKind,
    TextDocumentSyncOptions,
    TextEdit,
    VersionedTextDocumentIdentifier,
    WatchKind,
    WillSaveTextDocumentParams,
)
from ..text_document import TextDocument

if TYPE_CHECKING:
    from ..protocol import LanguageServerProtocol

from .protocol_part import LanguageServerProtocolPart

__all__ = ["TextDocumentProtocolPart", "LanguageServerDocumentException"]


class LanguageServerDocumentException(JsonRPCException):
    pass


class TextDocumentProtocolPart(LanguageServerProtocolPart):

    _logger = LoggingDescriptor()

    def __init__(self, parent: LanguageServerProtocol) -> None:
        super().__init__(parent)
        self._documents: Dict[DocumentUri, TextDocument] = {}
        self.parent.on_initialized.add(self._protocol_initialized)
        self._lock = Lock()

    async def _protocol_initialized(self, sender: Any) -> None:
        await self._update_filewatchers()

    async def _update_filewatchers(self) -> None:
        if self.parent.file_extensions:
            await self.parent.workspace.add_file_watcher(
                self._file_watcher,
                f"**/*.{{{','.join(self.parent.file_extensions)}}}",
                WatchKind.CHANGE | WatchKind.DELETE,
            )

    async def _file_watcher(self, sender: Any, changes: List[FileEvent]) -> None:
        to_change: Dict[str, FileEvent] = {}
        for change in changes:
            to_change[change.uri] = change

        for change in to_change.values():
            document = self._documents.get(DocumentUri(Uri(change.uri).normalized()), None)
            if document is not None and not document.opened_in_editor:
                await self.close_document(document, True)
                await self.did_close(self, document, callback_filter=language_id_filter(document))

    @async_tasking_event
    async def did_open(sender, document: TextDocument) -> None:  # NOSONAR
        ...

    @async_tasking_event
    async def did_close(sender, document: TextDocument) -> None:  # NOSONAR
        ...

    @async_tasking_event
    async def did_change(sender, document: TextDocument) -> None:  # NOSONAR
        ...

    @async_tasking_event
    async def did_save(sender, document: TextDocument) -> None:  # NOSONAR
        ...

    async def get(self, _uri: Union[DocumentUri, Uri]) -> Optional[TextDocument]:
        async with self._lock:
            return self._documents.get(str(Uri(_uri).normalized() if not isinstance(_uri, Uri) else _uri), None)

    def __len__(self) -> int:
        return self._documents.__len__()

    def __iter__(self) -> Iterator[DocumentUri]:
        return self._documents.__iter__()

    def __hash__(self) -> int:
        return id(self)

    async def _create_document(
        self,
        document_uri: DocumentUri,
        text: str,
        language_id: Optional[str] = None,
        version: Optional[int] = None,
    ) -> TextDocument:

        return TextDocument(
            document_uri=document_uri,
            language_id=language_id,
            text=text,
            version=version,
        )

    async def append_document(
        self,
        document_uri: DocumentUri,
        language_id: str,
        text: str,
        version: Optional[int] = None,
    ) -> TextDocument:

        async with self._lock:
            document = await self._create_document(
                document_uri=document_uri, language_id=language_id, text=text, version=version
            )

            self._documents[document_uri] = document

            return document

    __NORMALIZE_LINE_ENDINGS = re.compile(r"(\r?\n)")

    @classmethod
    def _normalize_line_endings(cls, text: str) -> str:
        return cls.__NORMALIZE_LINE_ENDINGS.sub("\n", text)

    @rpc_method(name="textDocument/didOpen", param_type=DidOpenTextDocumentParams)
    @_logger.call
    async def _text_document_did_open(self, text_document: TextDocumentItem, *args: Any, **kwargs: Any) -> None:
        async with self._lock:
            uri = str(Uri(text_document.uri).normalized())
            document = self._documents.get(uri, None)

            text_changed = True
            normalized_text = self._normalize_line_endings(text_document.text)

            if document is None:
                document = await self._create_document(
                    text_document.uri, normalized_text, text_document.language_id, text_document.version
                )

                self._documents[uri] = document
            else:
                text_changed = await document.text() != normalized_text
                if text_changed:
                    await document.apply_full_change(text_document.version, normalized_text)

            document.opened_in_editor = True
            document.references.add(self)

        await self.did_open(self, document, callback_filter=language_id_filter(document))

        if text_changed:
            await self.did_change(self, document, callback_filter=language_id_filter(document))

    @rpc_method(name="textDocument/didClose", param_type=DidCloseTextDocumentParams)
    @_logger.call
    async def _text_document_did_close(self, text_document: TextDocumentIdentifier, *args: Any, **kwargs: Any) -> None:
        uri = str(Uri(text_document.uri).normalized())
        document = await self.get(uri)

        if document is not None:
            document.references.remove(self)
            document.opened_in_editor = False

            await self.close_document(document)
            await self.did_close(self, document, callback_filter=language_id_filter(document))

    @_logger.call
    async def close_document(self, document: TextDocument, ignore_references: bool = False) -> None:
        async with self._lock:
            if len(document.references) == 0 or ignore_references:
                self._documents.pop(str(document.uri), None)

                await document.clear()
                del document
            else:
                document._version = None
                if await document.revert(None):
                    await self.did_change(self, document, callback_filter=language_id_filter(document))

            gc.collect()

    @rpc_method(name="textDocument/willSave", param_type=WillSaveTextDocumentParams)
    @_logger.call
    async def _text_document_will_save(
        self, text_document: TextDocumentIdentifier, reason: TextDocumentSaveReason, *args: Any, **kwargs: Any
    ) -> None:
        # TODO: implement
        pass

    @rpc_method(name="textDocument/didSave", param_type=DidSaveTextDocumentParams)
    @_logger.call
    async def _text_document_did_save(
        self, text_document: TextDocumentIdentifier, text: Optional[str] = None, *args: Any, **kwargs: Any
    ) -> None:
        document = await self.get(str(Uri(text_document.uri).normalized()))
        self._logger.warning(lambda: f"Document {text_document.uri} is not opened.", condition=lambda: document is None)

        if document is not None:
            if text is not None:
                normalized_text = self._normalize_line_endings(text)

                text_changed = await document.text() != normalized_text
                if text_changed:
                    await document.save(None, text)
                    await self.did_change(self, document, callback_filter=language_id_filter(document))

            await self.did_save(self, document, callback_filter=language_id_filter(document))

    @rpc_method(name="textDocument/willSaveWaitUntil", param_type=WillSaveTextDocumentParams)
    @_logger.call
    async def _text_document_will_save_wait_until(
        self, text_document: TextDocumentIdentifier, reason: TextDocumentSaveReason, *args: Any, **kwargs: Any
    ) -> List[TextEdit]:
        return []

    @rpc_method(name="textDocument/didChange", param_type=DidChangeTextDocumentParams)
    @_logger.call
    async def _text_document_did_change(
        self,
        text_document: VersionedTextDocumentIdentifier,
        content_changes: List[TextDocumentContentChangeEvent],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        document = await self.get(str(Uri(text_document.uri).normalized()))
        if document is None:
            raise LanguageServerDocumentException(f"Document {text_document.uri} is not opened.")

        sync_kind = (
            self.parent.capabilities.text_document_sync
            if isinstance(self.parent.capabilities.text_document_sync, TextDocumentSyncKind)
            else self.parent.capabilities.text_document_sync.change
            if isinstance(self.parent.capabilities.text_document_sync, TextDocumentSyncOptions)
            else None
        )
        for content_change in content_changes:
            if sync_kind is None or sync_kind == TextDocumentSyncKind.NONE:
                await document.apply_none_change()
            elif sync_kind == TextDocumentSyncKind.FULL and isinstance(
                content_change, TextDocumentContentTextChangeEvent
            ):
                await document.apply_full_change(
                    text_document.version, self._normalize_line_endings(content_change.text)
                )
            elif sync_kind == TextDocumentSyncKind.INCREMENTAL and isinstance(
                content_change, TextDocumentContentRangeChangeEvent
            ):
                await document.apply_incremental_change(
                    text_document.version, content_change.range, self._normalize_line_endings(content_change.text)
                )
            else:
                raise LanguageServerDocumentException(
                    f"Invalid type for content_changes {type(content_change)} "
                    f"and server capability {self.parent.capabilities.text_document_sync} "
                    f"for document {text_document.uri}."
                )

        await self.did_change(self, document, callback_filter=language_id_filter(document))



from __future__ import annotations

import ast
from typing import TYPE_CHECKING, Any, List, Optional, cast

from ....utils.async_tools import threaded
from ....utils.logging import LoggingDescriptor
from ...common.decorators import language_id
from ...common.lsp_types import FoldingRange
from ...common.text_document import TextDocument

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol  # pragma: no cover

from .protocol_part import RobotLanguageServerProtocolPart


class RobotFoldingRangeProtocolPart(RobotLanguageServerProtocolPart):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        parent.folding_ranges.collect.add(self.collect)

    @language_id("robotframework")
    @threaded()
    @_logger.call
    async def collect(self, sender: Any, document: TextDocument) -> Optional[List[FoldingRange]]:

        from ..utils.async_ast import AsyncVisitor

        class Visitor(AsyncVisitor):
            def __init__(self, parent: RobotFoldingRangeProtocolPart) -> None:
                super().__init__()
                self.parent = parent
                self.line_folding_only = True
                if (
                    self.parent.parent.client_capabilities
                    and self.parent.parent.client_capabilities.text_document
                    and self.parent.parent.client_capabilities.text_document.folding_range
                    and self.parent.parent.client_capabilities.text_document.folding_range.line_folding_only is not None
                ):
                    self.line_folding_only = (
                        self.parent.parent.client_capabilities.text_document.folding_range.line_folding_only
                    )

                self.result: List[FoldingRange] = []

            async def visit(self, node: ast.AST) -> None:
                await super().visit(node)

            @classmethod
            async def find_from(
                cls, model: ast.AST, parent: RobotFoldingRangeProtocolPart
            ) -> Optional[List[FoldingRange]]:

                finder = cls(parent)

                await finder.visit(model)

                return finder.result if finder.result else None

            def __append(self, node: ast.AST, kind: str) -> None:
                if not self.line_folding_only:
                    self.result.append(
                        FoldingRange(
                            start_line=node.lineno - 1,
                            end_line=node.end_lineno - 1 if node.end_lineno is not None else node.lineno - 1,
                            start_character=node.col_offset if not self.line_folding_only else None,
                            end_character=node.end_col_offset if not self.line_folding_only else None,
                            kind=kind,
                        )
                    )
                else:
                    self.result.append(
                        FoldingRange(
                            start_line=node.lineno - 1,
                            end_line=node.end_lineno - 1 if node.end_lineno is not None else node.lineno - 1,
                            kind=kind,
                        )
                    )

            async def visit_Section(self, node: ast.AST) -> None:  # noqa: N802
                self.__append(node, kind="section")

                await self.generic_visit(node)

            async def visit_CommentSection(self, node: ast.AST) -> None:  # noqa: N802
                self.__append(node, kind="comment")
                await self.generic_visit(node)

            async def visit_TestCase(self, node: ast.AST) -> None:  # noqa: N802
                from robot.parsing.model.blocks import TestCase

                if cast(TestCase, node).name:
                    self.__append(node, kind="testcase")
                    await self.generic_visit(node)

            async def visit_Keyword(self, node: ast.AST) -> None:  # noqa: N802
                from robot.parsing.model.blocks import Keyword

                if cast(Keyword, node).name:
                    self.__append(node, kind="keyword")
                    await self.generic_visit(node)

            async def visit_ForLoop(self, node: ast.AST) -> None:  # noqa: N802, pragma: no cover
                self.__append(node, kind="for_loop")
                await self.generic_visit(node)

            async def visit_For(self, node: ast.AST) -> None:  # noqa: N802
                self.__append(node, kind="for")
                await self.generic_visit(node)

            async def visit_If(self, node: ast.AST) -> None:  # noqa: N802
                self.__append(node, kind="if")
                await self.generic_visit(node)

        return await Visitor.find_from(await self.parent.documents_cache.get_model(document), self)



from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, Tuple

from ...common.lsp_types import Position, Range
from ..utils.ast import Token


@dataclass
class SourceEntity:
    line_no: int
    col_offset: int
    end_line_no: int
    end_col_offset: int
    source: str


@dataclass
class Import(SourceEntity):
    name: Optional[str]
    name_token: Optional[Token]

    def range(self) -> Range:
        return Range(
            start=Position(
                line=self.name_token.lineno - 1 if self.name_token is not None else self.line_no - 1,
                character=self.name_token.col_offset if self.name_token is not None else self.col_offset,
            ),
            end=Position(
                line=self.name_token.lineno - 1 if self.name_token is not None else self.end_line_no - 1,
                character=self.name_token.end_col_offset if self.name_token is not None else self.end_col_offset,
            ),
        )


@dataclass
class LibraryImport(Import):
    args: Tuple[str, ...] = ()
    alias: Optional[str] = None

    def __hash__(self) -> int:
        return hash(
            (
                type(self),
                self.name,
                self.args,
                self.alias,
            )
        )


@dataclass
class ResourceImport(Import):
    def __hash__(self) -> int:
        return hash(
            (
                type(self),
                self.name,
            )
        )


@dataclass
class VariablesImport(Import):
    args: Tuple[str, ...] = ()

    def __hash__(self) -> int:
        return hash(
            (
                type(self),
                self.name,
                self.args,
            )
        )


class VariableDefinitionType(Enum):
    VARIABLE = "variable"
    LOCAL_VARIABLE = "local variable"
    ARGUMENT = "argument"
    COMMAND_LINE_VARIABLE = "command line variable"
    BUILTIN_VARIABLE = "builtin variable"
    IMPORTED_VARIABLE = "imported variable"
    ENVIRONMENT_VARIABLE = "environment variable"


@dataclass
class VariableDefinition(SourceEntity):
    name: str
    name_token: Optional[Token]
    type: VariableDefinitionType = VariableDefinitionType.VARIABLE

    has_value: bool = False
    resolvable: bool = False

    value: Any = None

    def __hash__(self) -> int:
        return hash((type(self), self.name, self.type, self.range, self.source, self.name_token))

    def range(self) -> Range:
        return Range(
            start=Position(
                line=self.line_no - 1,
                character=self.col_offset,
            ),
            end=Position(
                line=self.end_line_no - 1,
                character=self.end_col_offset,
            ),
        )


@dataclass
class LocalVariableDefinition(VariableDefinition):
    type: VariableDefinitionType = VariableDefinitionType.LOCAL_VARIABLE

    def __hash__(self) -> int:
        return hash((type(self), self.name, self.type))


@dataclass
class BuiltInVariableDefinition(VariableDefinition):
    type: VariableDefinitionType = VariableDefinitionType.BUILTIN_VARIABLE
    resolvable: bool = True

    def __hash__(self) -> int:
        return hash((type(self), self.name, self.type))


@dataclass
class CommandLineVariableDefinition(VariableDefinition):
    type: VariableDefinitionType = VariableDefinitionType.COMMAND_LINE_VARIABLE
    resolvable: bool = True

    def __hash__(self) -> int:
        return hash((type(self), self.name, self.type))


@dataclass
class ArgumentDefinition(VariableDefinition):
    type: VariableDefinitionType = VariableDefinitionType.ARGUMENT

    def __hash__(self) -> int:
        return hash((type(self), self.name, self.type))


@dataclass
class ImportedVariableDefinition(VariableDefinition):
    type: VariableDefinitionType = VariableDefinitionType.IMPORTED_VARIABLE

    def __hash__(self) -> int:
        return hash((type(self), self.name, self.type))


@dataclass
class EnvironmentVariableDefinition(VariableDefinition):
    type: VariableDefinitionType = VariableDefinitionType.ENVIRONMENT_VARIABLE
    resolvable: bool = True

    def __hash__(self) -> int:
        return hash((type(self), self.name, self.type))



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

from ..common.parts.workspace import ConfigBase, config_section


@config_section("robotcode.languageServer")
@dataclass
class LanguageServerConfig(ConfigBase):
    mode: str = "stdio"
    tcp_port: int = 0
    args: Tuple[str, ...] = field(default_factory=tuple)


@config_section("robotcode.robot")
@dataclass
class RobotConfig(ConfigBase):
    args: Tuple[str, ...] = field(default_factory=tuple)
    python_path: List[str] = field(default_factory=list)
    env: Dict[str, str] = field(default_factory=dict)
    variables: Dict[str, Any] = field(default_factory=dict)
    output_dir: Optional[str] = None
    output_file: Optional[str] = None
    log_file: Optional[str] = None
    debug_file: Optional[str] = None
    log_level: Optional[str] = None


@config_section("robotcode.syntax")
@dataclass
class SyntaxConfig(ConfigBase):
    section_style: str = "*** {name}s ***"


@config_section("robotcode.robocop")
@dataclass
class RoboCopConfig(ConfigBase):
    enabled: bool = True
    include: List[str] = field(default_factory=list)
    exclude: List[str] = field(default_factory=list)
    configurations: List[str] = field(default_factory=list)


@config_section("robotcode.robotidy")
@dataclass
class RoboTidyConfig(ConfigBase):
    enabled: bool = True


@config_section("robotcode.workspace")
@dataclass
class WorkspaceConfig(ConfigBase):
    exclude_patterns: List[str] = field(default_factory=list)


@config_section("robotcode")
@dataclass
class RobotCodeConfig(ConfigBase):
    language_server: LanguageServerConfig = field(default_factory=LanguageServerConfig)
    robot: RobotConfig = field(default_factory=RobotConfig)
    syntax: SyntaxConfig = field(default_factory=SyntaxConfig)
    workspace: WorkspaceConfig = field(default_factory=WorkspaceConfig)



from __future__ import annotations

import asyncio
import io
import os
from typing import TYPE_CHECKING, Any, List, Optional, cast

from ....utils.async_tools import threaded
from ....utils.logging import LoggingDescriptor
from ...common.decorators import language_id
from ...common.lsp_types import (
    FormattingOptions,
    MessageType,
    Position,
    Range,
    TextEdit,
)
from ...common.text_document import TextDocument
from ..utils.version import get_robot_version

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol

from ..configuration import RoboTidyConfig
from .model_helper import ModelHelperMixin
from .protocol_part import RobotLanguageServerProtocolPart


def robotidy_installed() -> bool:
    try:
        __import__("robotidy")
    except ImportError:
        return False
    return True


class RobotFormattingProtocolPart(RobotLanguageServerProtocolPart, ModelHelperMixin):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        parent.formatting.format.add(self.format)

        # TODO implement range formatting
        # parent.formatting.format_range.add(self.format_range)

        self.space_count = 4
        self.use_pipes = False
        self.line_separator = os.linesep
        self.short_test_name_length = 18
        self.setting_and_variable_name_length = 14

    async def get_config(self, document: TextDocument) -> Optional[RoboTidyConfig]:
        folder = self.parent.workspace.get_workspace_folder(document.uri)
        if folder is None:
            return None

        return await self.parent.workspace.get_configuration(RoboTidyConfig, folder.uri)

    @language_id("robotframework")
    @threaded()
    @_logger.call
    async def format(
        self, sender: Any, document: TextDocument, options: FormattingOptions, **further_options: Any
    ) -> Optional[List[TextEdit]]:
        config = await self.get_config(document)

        if config and (config.enabled or get_robot_version() >= (5, 0)) and robotidy_installed():
            return await self.format_robot_tidy(document, options, **further_options)

        if get_robot_version() < (5, 0):
            return await self.format_internal(document, options, **further_options)

        self.parent.window.show_message(
            "RobotFramework formatter is not available, please install 'robotframework-tidy'.", MessageType.ERROR
        )

        return None

    async def format_robot_tidy(
        self, document: TextDocument, options: FormattingOptions, **further_options: Any
    ) -> Optional[List[TextEdit]]:

        from difflib import SequenceMatcher

        from robotidy.api import RobotidyAPI

        try:
            model = await self.parent.documents_cache.get_model(document)

            robot_tidy = RobotidyAPI(document.uri.to_path(), None)

            changed, _, new = robot_tidy.transform(model)

            if not changed:
                return None

            new_lines = new.text.splitlines()

            result: List[TextEdit] = []
            matcher = SequenceMatcher(a=await document.get_lines(), b=new_lines, autojunk=False)
            for code, old_start, old_end, new_start, new_end in matcher.get_opcodes():
                if code == "insert" or code == "replace":
                    result.append(
                        TextEdit(
                            range=Range(
                                start=Position(line=old_start, character=0),
                                end=Position(line=old_end, character=0),
                            ),
                            new_text=os.linesep.join(new_lines[new_start:new_end]) + os.linesep,
                        )
                    )

                elif code == "delete":
                    result.append(
                        TextEdit(
                            range=Range(
                                start=Position(line=old_start, character=0),
                                end=Position(line=old_end, character=0),
                            ),
                            new_text="",
                        )
                    )

            if result:
                return result

        except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
            raise
        except BaseException as e:
            self._logger.exception(e)
        return None

    async def format_internal(
        self, document: TextDocument, options: FormattingOptions, **further_options: Any
    ) -> Optional[List[TextEdit]]:

        from robot.parsing.model.blocks import File
        from robot.tidypkg import (
            Aligner,
            Cleaner,
            NewlineNormalizer,
            SeparatorNormalizer,
        )

        model = cast(File, await self.parent.documents_cache.get_model(document))

        Cleaner().visit(model)
        NewlineNormalizer(self.line_separator, self.short_test_name_length).visit(model)
        SeparatorNormalizer(self.use_pipes, self.space_count).visit(model)
        Aligner(self.short_test_name_length, self.setting_and_variable_name_length, self.use_pipes).visit(model)

        with io.StringIO() as s:
            model.save(s)

            return [
                TextEdit(
                    range=Range(
                        start=Position(line=0, character=0),
                        end=Position(
                            line=len(await document.get_lines()), character=len((await document.get_lines())[-1])
                        ),
                    ),
                    new_text=s.getvalue(),
                )
            ]

    @language_id("robotframework")
    async def format_range(
        self, sender: Any, document: TextDocument, range: Range, options: FormattingOptions, **further_options: Any
    ) -> Optional[List[TextEdit]]:
        # TODO implement range formatting
        # config = await self.get_config(document)
        # if config and config.enabled and robotidy_installed():
        #     return await self.format_robot_tidy(document, options, range=range, **further_options)
        return None



from __future__ import annotations

import ast
import io
import weakref
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Iterable,
    List,
    Optional,
    cast,
)

from ....utils.async_tools import (
    Lock,
    async_tasking_event,
    check_canceled_sync,
    create_sub_task,
)
from ....utils.uri import Uri
from ...common.decorators import language_id_filter
from ...common.parts.workspace import WorkspaceFolder
from ...common.text_document import TextDocument
from ..configuration import RobotConfig
from ..diagnostics.imports_manager import ImportsManager
from ..diagnostics.namespace import DocumentType, Namespace
from ..utils.ast import Token

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol

from .protocol_part import RobotLanguageServerProtocolPart


class UnknownFileTypeError(Exception):
    pass


class DocumentsCache(RobotLanguageServerProtocolPart):
    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        self._imports_managers_lock = Lock()
        self._imports_managers: weakref.WeakKeyDictionary[WorkspaceFolder, ImportsManager] = weakref.WeakKeyDictionary()
        self._default_imports_manager: Optional[ImportsManager] = None

    async def get_document_type(self, document: TextDocument) -> DocumentType:
        return await document.get_cache(self.__get_document_type)

    async def __get_document_type(self, document: TextDocument) -> DocumentType:
        path = document.uri.to_path()
        suffix = path.suffix.lower()

        if path.name == "__init__.robot":
            return DocumentType.INIT
        elif suffix == ".robot":
            return DocumentType.GENERAL
        elif suffix == ".resource":
            return DocumentType.RESOURCE
        else:
            return DocumentType.UNKNOWN

    async def get_tokens(self, document: TextDocument) -> List[Token]:
        return await document.get_cache(self.__get_tokens)

    async def __get_tokens(self, document: TextDocument) -> List[Token]:
        document_type = await self.get_document_type(document)
        if document_type == DocumentType.INIT:
            return await self.get_init_tokens(document)
        elif document_type == DocumentType.GENERAL:
            return await self.get_general_tokens(document)
        elif document_type == DocumentType.RESOURCE:
            return await self.get_resource_tokens(document)
        else:
            raise UnknownFileTypeError(str(document.uri))

    async def get_general_tokens(self, document: TextDocument) -> List[Token]:
        return await document.get_cache(self.__get_general_tokens)

    async def __get_general_tokens(self, document: TextDocument) -> List[Token]:
        import robot.api

        def get(text: str) -> List[Token]:
            with io.StringIO(text) as content:
                return [e for e in robot.api.get_tokens(content) if check_canceled_sync()]

        return await self.__get_tokens_internal(document, get)

    async def __get_tokens_internal(
        self,
        document: TextDocument,
        get: Callable[[str], List[Token]],
    ) -> List[Token]:

        return get(await document.text())

    async def get_resource_tokens(self, document: TextDocument) -> List[Token]:
        return await document.get_cache(self.__get_resource_tokens)

    async def __get_resource_tokens(self, document: TextDocument) -> List[Token]:
        import robot.api

        def get(text: str) -> List[Token]:
            with io.StringIO(text) as content:
                return [e for e in robot.api.get_resource_tokens(content) if check_canceled_sync()]

        return await self.__get_tokens_internal(document, get)

    async def get_init_tokens(self, document: TextDocument) -> List[Token]:
        return await document.get_cache(self.__get_init_tokens)

    async def __get_init_tokens(self, document: TextDocument) -> List[Token]:
        import robot.api

        def get(text: str) -> List[Token]:
            with io.StringIO(text) as content:
                return [e for e in robot.api.get_init_tokens(content) if check_canceled_sync()]

        return await self.__get_tokens_internal(document, get)

    async def get_model(self, document: TextDocument) -> ast.AST:
        document_type = await self.get_document_type(document)

        if document_type == DocumentType.INIT:
            return await self.get_init_model(document)
        if document_type == DocumentType.GENERAL:
            return await self.get_general_model(document)
        if document_type == DocumentType.RESOURCE:
            return await self.get_resource_model(document)
        else:
            raise UnknownFileTypeError(f"Unknown file type '{document.uri}'.")

    def __get_model(self, document: TextDocument, tokens: Iterable[Any], document_type: DocumentType) -> ast.AST:
        from robot.parsing.lexer import Token
        from robot.parsing.parser.parser import _get_model

        def get_tokens(_source: str, _data_only: bool = False) -> Generator[Token, None, None]:
            for t in tokens:
                check_canceled_sync()

                yield t

        model = _get_model(get_tokens, document.uri.to_path())

        setattr(model, "source", str(document.uri.to_path()))
        setattr(model, "model_type", document_type)

        return cast(ast.AST, model)

    async def get_general_model(self, document: TextDocument) -> ast.AST:
        return await document.get_cache(self.__get_general_model)

    async def __get_general_model(self, document: TextDocument) -> ast.AST:
        return self.__get_model(document, await self.get_general_tokens(document), DocumentType.GENERAL)

    async def get_resource_model(self, document: TextDocument) -> ast.AST:
        return await document.get_cache(self.__get_resource_model)

    async def __get_resource_model(self, document: TextDocument) -> ast.AST:
        return self.__get_model(document, await self.get_resource_tokens(document), DocumentType.RESOURCE)

    async def get_init_model(self, document: TextDocument) -> ast.AST:
        return await document.get_cache(self.__get_init_model)

    async def __get_init_model(self, document: TextDocument) -> ast.AST:
        return self.__get_model(document, await self.get_init_tokens(document), DocumentType.INIT)

    async def get_namespace(self, document: TextDocument) -> Namespace:
        return await document.get_cache(self.__get_namespace)

    async def __get_namespace(self, document: TextDocument) -> Namespace:
        return await self.__get_namespace_for_document_type(document, None)

    async def get_resource_namespace(self, document: TextDocument) -> Namespace:
        return await document.get_cache(self.__get_resource_namespace)

    async def __get_resource_namespace(self, document: TextDocument) -> Namespace:
        return await self.__get_namespace_for_document_type(document, DocumentType.RESOURCE)

    async def get_init_namespace(self, document: TextDocument) -> Namespace:
        return await document.get_cache(self.__get_init_namespace)

    async def __get_init_namespace(self, document: TextDocument) -> Namespace:
        return await self.__get_namespace_for_document_type(document, DocumentType.INIT)

    async def get_general_namespace(self, document: TextDocument) -> Namespace:
        return await document.get_cache(self.__get_general_namespace)

    async def __get_general_namespace(self, document: TextDocument) -> Namespace:
        return await self.__get_namespace_for_document_type(document, DocumentType.GENERAL)

    @async_tasking_event
    async def namespace_invalidated(sender, document: TextDocument) -> None:  # NOSONAR
        ...

    async def __invalidate_namespace(self, namespace: Namespace) -> None:
        document = namespace.document
        if document is not None:
            await document.remove_cache_entry(self.__get_namespace)
            await self.namespace_invalidated(
                self,
                document,
                callback_filter=language_id_filter(document),
            )

    async def __get_namespace_for_document_type(
        self,
        document: TextDocument,
        document_type: Optional[DocumentType],
    ) -> Namespace:
        if document_type is not None and document_type == DocumentType.INIT:
            model = await self.get_init_model(document)
        elif document_type is not None and document_type == DocumentType.RESOURCE:
            model = await self.get_resource_model(document)
        elif document_type is not None and document_type == DocumentType.GENERAL:
            model = await self.get_general_model(document)
        else:
            model = await self.get_model(document)

        imports_manager = await self.get_imports_manager(document)

        def invalidate(namespace: Namespace) -> None:
            create_sub_task(self.__invalidate_namespace(namespace))

        return Namespace(imports_manager, model, str(document.uri.to_path()), invalidate, document, document_type)

    @property
    async def default_imports_manager(self) -> ImportsManager:
        if self._default_imports_manager is None:
            async with self._imports_managers_lock:
                if self._default_imports_manager is None:
                    self._default_imports_manager = ImportsManager(
                        self.parent,
                        Uri(self.parent.workspace.root_uri or "."),
                        RobotConfig(args=(), python_path=[], env={}, variables={}),
                    )
        return self._default_imports_manager

    async def get_imports_manager(self, document: TextDocument) -> ImportsManager:
        folder = self.parent.workspace.get_workspace_folder(document.uri)
        if folder is None:
            return await self.default_imports_manager

        if folder not in self._imports_managers:
            async with self._imports_managers_lock:
                if folder not in self._imports_managers:
                    config = await self.parent.workspace.get_configuration(RobotConfig, folder.uri)

                    self._imports_managers[folder] = ImportsManager(self.parent, folder.uri, config)
        return self._imports_managers[folder]



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, Iterator, List, Literal, Optional, Union

from ..utils.dataclasses import to_camel_case, to_snake_case


@dataclass
class Model:
    @classmethod
    def _encode_case(cls, s: str) -> str:
        return to_camel_case(s)

    @classmethod
    def _decode_case(cls, s: str) -> str:
        return to_snake_case(s)


def __next_id_iter() -> Iterator[int]:
    i = 0
    while True:
        yield i
        i += 1


_next_id_iterator = __next_id_iter()


def _next_id() -> int:
    return next(_next_id_iterator)


@dataclass
class ProtocolMessage(Model):
    type: Union[Literal["request", "response", "event"], str]
    seq: int = field(default_factory=lambda: _next_id())


@dataclass
class _Request(Model):
    command: str
    arguments: Optional[Any] = None


@dataclass
class Request(ProtocolMessage, _Request):
    type: str = "request"


@dataclass
class _Event(Model):
    event: str


@dataclass
class Event(ProtocolMessage, _Event):
    type: str = "event"
    body: Optional[Any] = None


@dataclass
class _Response(Model):
    request_seq: int = field(metadata={"alias": "request_seq"})
    success: bool
    command: str
    message: Optional[Union[Literal["cancelled"], str]] = None


@dataclass
class Response(ProtocolMessage, _Response):
    type: str = "response"
    body: Optional[Any] = None


@dataclass
class Message(Model):
    format: str
    id: int = -1
    variables: Optional[Dict[str, str]] = None
    send_telemetry: Optional[bool] = None
    show_user: Optional[bool] = None
    url: Optional[str] = None
    url_label: Optional[str] = None

    def __str__(self) -> str:
        result = self.format

        for k, v in (self.variables or {}).items():
            result = result.replace(f"{{{k}}}", v)

        result += f" ({super().__str__()})"
        return result


@dataclass
class ErrorBody(Model):
    error: Optional[Message] = None


@dataclass
class _ErrorResponse(Model):
    body: Optional[ErrorBody]


@dataclass
class ErrorResponse(Response, _ErrorResponse):
    body: Optional[ErrorBody] = field()


@dataclass
class CancelArguments(Model):
    request_id: Optional[int] = None
    progress_id: Optional[int] = None


@dataclass
class CancelRequest(Request):
    arguments: Optional[CancelArguments] = None
    command: str = "cancel"


@dataclass
class CancelResponse(Response):
    pass


@dataclass
class InitializedEvent(Event):
    event: str = "initialized"


class StoppedReason(Enum):
    STEP = "step"
    BREAKPOINT = "breakpoint"
    EXCEPTION = "exception"
    PAUSE = "pause"
    ENTRY = "entry"
    GOTO = "goto"
    FUNCTION_BREAKPOINT = "function breakpoint"
    DATA_BREAKPOINT = "data breakpoint"
    INSTRUCTION_BREAKPOINT = "instruction breakpoint"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass
class StoppedEventBody(Model):
    reason: Union[
        StoppedReason,
        str,
    ]
    description: Optional[str] = None
    thread_id: Optional[int] = None
    preserve_focus_hint: Optional[bool] = None
    text: Optional[str] = None
    all_threads_stopped: Optional[bool] = None
    hit_breakpoint_ids: Optional[List[int]] = None


@dataclass
class _StoppedEvent(Model):
    body: StoppedEventBody


@dataclass
class StoppedEvent(Event, _StoppedEvent):
    body: StoppedEventBody = field()
    event: str = "stopped"


@dataclass
class ContinuedEventBody(Model):
    thread_id: int
    all_threads_continued: Optional[bool]


@dataclass
class _ContinuedEvent(Model):
    body: ContinuedEventBody


@dataclass
class ContinuedEvent(Event, _ContinuedEvent):
    body: ContinuedEventBody = field()
    event: str = "continued"


@dataclass
class ExitedEventBody(Model):
    exit_code: int


@dataclass
class _ExitedEvent(Model):
    body: ExitedEventBody


@dataclass
class ExitedEvent(Event, _ExitedEvent):
    body: ExitedEventBody = field()
    event: str = "exited"


@dataclass
class TerminatedEventBody(Model):
    restart: Optional[Any] = None


@dataclass
class _TerminatedEvent(Model):
    body: Optional[TerminatedEventBody] = None


@dataclass
class TerminatedEvent(Event, _TerminatedEvent):
    body: Optional[TerminatedEventBody] = None
    event: str = "terminated"


class ChecksumAlgorithm(Enum):
    MD5 = "MD5"
    SHA1 = "SHA1"
    SHA256 = "SHA256"
    TIMESTAMP = "timestamp"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass
class Checksum(Model):
    algorithm: ChecksumAlgorithm
    checksum: str


@dataclass
class Source(Model):
    name: Optional[str] = None
    path: Optional[str] = None
    source_reference: Optional[int] = None
    presentation_hint: Optional[Literal["normal", "emphasize", "deemphasize"]] = None
    origin: Optional[str] = None
    sources: Optional[List[Source]] = None
    adapter_data: Optional[Any] = None
    checksums: Optional[List[Checksum]] = None


class OutputCategory(Enum):
    CONSOLE = "console"
    STDOUT = "stdout"
    STDERR = "stderr"
    TELEMETRY = "telemetry"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


class OutputGroup(Enum):
    START = "start"
    STARTCOLLAPSED = "startCollapsed"
    END = "end"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass
class OutputEventBody(Model):
    output: str
    category: Union[OutputCategory, str, None] = None
    group: Optional[OutputGroup] = None
    variables_reference: Optional[int] = None
    source: Optional[Source] = None
    line: Optional[int] = None
    column: Optional[int] = None
    data: Optional[Any] = None


@dataclass
class _OutputEvent(Model):
    body: Optional[OutputEventBody] = None


@dataclass
class OutputEvent(Event, _OutputEvent):
    body: Optional[OutputEventBody] = None
    event: str = "output"


@dataclass
class InitializeRequestArguments(Model):
    adapter_id: str = field(metadata={"alias": "adapterID"})
    client_id: Optional[str] = field(metadata={"alias": "clientID"})
    client_name: Optional[str] = None
    locale: Optional[str] = None
    lines_start_at1: Optional[bool] = None
    columns_start_at1: Optional[bool] = None
    path_format: Optional[Union[Literal["path", "uri"], str]] = None
    supports_variable_type: Optional[bool] = None
    supports_variable_paging: Optional[bool] = None
    supports_run_in_terminal_request: Optional[bool] = None
    supports_memory_references: Optional[bool] = None
    supports_progress_reporting: Optional[bool] = None
    supports_invalidated_event: Optional[bool] = None


@dataclass
class _InitializeRequest(Model):
    arguments: InitializeRequestArguments


@dataclass
class InitializeRequest(Request, _InitializeRequest):
    arguments: InitializeRequestArguments = field()
    command: str = "initialize"


@dataclass
class ExceptionBreakpointsFilter(Model):
    filter: str
    label: str
    description: Optional[str] = None
    default: Optional[bool] = None
    supports_condition: Optional[bool] = None
    condition_description: Optional[str] = None


@dataclass
class ColumnDescriptor(Model):
    attribute_name: str
    label: str
    format: Optional[str] = None
    type: Optional[Literal["string", "number", "boolean", "unixTimestampUTC"]] = None
    width: Optional[int] = None


@dataclass
class Capabilities(Model):
    supports_configuration_done_request: Optional[bool] = None
    supports_function_breakpoints: Optional[bool] = None
    supports_conditional_breakpoints: Optional[bool] = None
    supports_hit_conditional_breakpoints: Optional[bool] = None
    supports_evaluate_for_hovers: Optional[bool] = None
    exception_breakpoint_filters: Optional[List[ExceptionBreakpointsFilter]] = None
    supports_step_back: Optional[bool] = None
    supports_set_variable: Optional[bool] = None
    supports_restart_frame: Optional[bool] = None
    supports_goto_targets_request: Optional[bool] = None
    supports_step_in_targets_request: Optional[bool] = None
    supports_completions_request: Optional[bool] = None
    completion_trigger_characters: Optional[List[str]] = None
    supports_modules_request: Optional[bool] = None
    additional_module_columns: Optional[List[ColumnDescriptor]] = None
    supported_checksum_algorithms: Optional[List[ChecksumAlgorithm]] = None
    supports_restart_request: Optional[bool] = None
    supports_exception_options: Optional[bool] = None
    supports_value_formatting_options: Optional[bool] = None
    supports_exception_info_request: Optional[bool] = None
    support_terminate_debuggee: Optional[bool] = None
    support_suspend_debuggee: Optional[bool] = None
    supports_delayed_stack_trace_loading: Optional[bool] = None
    supports_loaded_sources_request: Optional[bool] = None
    supports_log_points: Optional[bool] = None
    supports_terminate_threads_request: Optional[bool] = None
    supports_set_expression: Optional[bool] = None
    supports_terminate_request: Optional[bool] = None
    supports_data_breakpoints: Optional[bool] = None
    supports_read_memory_request: Optional[bool] = None
    supports_disassemble_request: Optional[bool] = None
    supports_cancel_request: Optional[bool] = None
    supports_breakpoint_locations_request: Optional[bool] = None
    supports_clipboard_context: Optional[bool] = None
    supports_stepping_granularity: Optional[bool] = None
    supports_instruction_breakpoints: Optional[bool] = None
    supports_exception_filter_options: Optional[bool] = None


@dataclass
class InitializeResponse(Response):
    body: Optional[Capabilities] = None


@dataclass
class LaunchRequestArguments(Model):
    no_debug: Optional[bool] = None
    restart: Optional[Any] = field(default=None, metadata={"alias": "__restart"})


@dataclass
class _LaunchRequest(Model):
    arguments: LaunchRequestArguments


@dataclass
class LaunchRequest(Request, _LaunchRequest):
    arguments: LaunchRequestArguments = field()
    command: str = "launch"


@dataclass
class LaunchResponse(Response):
    pass


class RunInTerminalKind(Enum):
    INTEGRATED = "integrated"
    EXTERNAL = "external"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass
class RunInTerminalRequestArguments(Model):
    cwd: str
    args: List[str]
    env: Optional[Dict[str, Optional[str]]] = None
    kind: Optional[RunInTerminalKind] = None
    title: Optional[str] = None


@dataclass
class _RunInTerminalRequest(Model):
    arguments: RunInTerminalRequestArguments


@dataclass
class RunInTerminalRequest(Request, _RunInTerminalRequest):
    arguments: RunInTerminalRequestArguments = field()
    command: str = field(default="runInTerminal", init=False)


@dataclass
class RunInTerminalResponseBody(Model):
    process_id: Optional[int] = None
    shell_process_id: Optional[int] = None


@dataclass
class _RunInTerminalResponse(Model):
    body: RunInTerminalResponseBody


@dataclass
class RunInTerminalResponse(Response, _RunInTerminalResponse):
    body: RunInTerminalResponseBody = field()


@dataclass
class ConfigurationDoneArguments(Model):
    pass


@dataclass
class _ConfigurationDoneRequest(Model):
    arguments: Optional[ConfigurationDoneArguments] = None


@dataclass
class ConfigurationDoneRequest(Request, _ConfigurationDoneRequest):
    arguments: Optional[ConfigurationDoneArguments] = None
    command: str = "configurationDone"


@dataclass
class ConfigurationDoneResponse(Response):
    pass


@dataclass
class DisconnectArguments(Model):
    restart: Optional[bool] = None
    terminate_debuggee: Optional[bool] = None
    suspend_debuggee: Optional[bool] = None


@dataclass
class _DisconnectRequest(Model):
    arguments: Optional[DisconnectArguments] = None


@dataclass
class DisconnectRequest(Request, _DisconnectRequest):
    arguments: Optional[DisconnectArguments] = None
    command: str = "disconnect"


@dataclass
class DisconnectResponse(Response):
    pass


@dataclass
class SourceBreakpoint(Model):
    line: int
    column: Optional[int] = None
    condition: Optional[str] = None
    hit_condition: Optional[str] = None
    log_message: Optional[str] = None


@dataclass
class SetBreakpointsArguments(Model):
    source: Source
    breakpoints: Optional[List[SourceBreakpoint]] = None
    lines: Optional[List[int]] = None
    source_modified: Optional[bool] = None


@dataclass
class _SetBreakpointsRequest(Model):
    arguments: SetBreakpointsArguments


@dataclass
class SetBreakpointsRequest(Request, _SetBreakpointsRequest):
    arguments: SetBreakpointsArguments = field()
    command: str = "setBreakpoints"


@dataclass
class Breakpoint(Model):
    verified: bool
    id: Optional[int] = None
    message: Optional[str] = None
    source: Optional[Source] = None
    line: Optional[int] = None
    column: Optional[int] = None
    end_line: Optional[int] = None
    end_column: Optional[int] = None
    instruction_reference: Optional[str] = None
    offset: Optional[int] = None


@dataclass
class SetBreakpointsResponseBody(Model):
    breakpoints: List[Breakpoint]


@dataclass
class _SetBreakpointsResponse(Model):
    body: SetBreakpointsResponseBody


@dataclass
class SetBreakpointsResponse(Response, _SetBreakpointsResponse):
    body: SetBreakpointsResponseBody = field()


@dataclass
class ThreadsRequest(Request):
    command: str = "threads"


@dataclass
class Thread(Model):
    id: int
    name: str


@dataclass
class ThreadsResponseBody(Model):
    threads: List[Thread]


@dataclass
class _ThreadsResponse(Model):
    body: ThreadsResponseBody


@dataclass
class ThreadsResponse(Response, _ThreadsResponse):
    body: ThreadsResponseBody = field()


@dataclass
class TerminateArguments(Model):
    restart: Optional[bool] = None


@dataclass
class _TerminateRequest(Model):
    arguments: Optional[TerminateArguments] = None


@dataclass
class TerminateRequest(Request, _TerminateRequest):
    arguments: Optional[TerminateArguments] = None
    command: str = "terminate"


@dataclass
class TerminateResponse(Response):
    pass


@dataclass
class StackFrameFormat(Model):
    parameters: Optional[bool] = None
    parameter_types: Optional[bool] = None
    parameter_names: Optional[bool] = None
    parameter_values: Optional[bool] = None
    line: Optional[bool] = None
    module: Optional[bool] = None
    include_all: Optional[bool] = None


@dataclass
class StackTraceArguments(Model):
    thread_id: int
    start_frame: Optional[int] = None
    levels: Optional[int] = None
    format: Optional[StackFrameFormat] = None


@dataclass
class _StackTraceRequest(Model):
    arguments: StackTraceArguments


@dataclass
class StackTraceRequest(Request, _StackTraceRequest):
    arguments: StackTraceArguments = field()
    command: str = "stackTrace"


@dataclass
class StackFrame(Model):
    id: int
    name: str
    line: int
    column: int
    source: Optional[Source] = None
    end_line: Optional[int] = None
    end_column: Optional[int] = None
    can_restart: Optional[bool] = None
    instruction_pointer_reference: Optional[str] = None
    module_id: Union[int, str, None] = None
    presentation_hint: Optional[Literal["normal", "label", "subtle"]] = None


@dataclass
class StackTraceResponseBody(Model):
    stack_frames: List[StackFrame]
    total_frames: Optional[int]


@dataclass
class _StackTraceResponse(Model):
    body: StackTraceResponseBody


@dataclass
class StackTraceResponse(Response, _StackTraceResponse):
    body: StackTraceResponseBody = field()


@dataclass
class ScopesArguments(Model):
    frame_id: int


@dataclass
class Scope(Model):
    name: str
    variables_reference: int
    expensive: bool
    presentation_hint: Union[Literal["arguments", "locals", "registers"], str, None] = None
    named_variables: Optional[int] = None
    indexed_variables: Optional[int] = None
    source: Optional[Source] = None
    line: Optional[int] = None
    column: Optional[int] = None
    end_line: Optional[int] = None
    end_column: Optional[int] = None


@dataclass
class _ScopesRequest(Model):
    arguments: ScopesArguments


@dataclass
class ScopesRequest(Request, _ScopesRequest):
    arguments: ScopesArguments = field()
    command: str = "scopes"


@dataclass
class ScopesResponseBody(Model):
    scopes: List[Scope]


@dataclass
class _ScopesResponse(Model):
    body: ScopesResponseBody


@dataclass
class ScopesResponse(Response, _ScopesResponse):
    body: ScopesResponseBody = field()


@dataclass
class ContinueArguments(Model):
    thread_id: int


@dataclass
class _ContinueRequest(Model):
    arguments: ContinueArguments


@dataclass
class ContinueRequest(Request, _ContinueRequest):
    arguments: ContinueArguments = field()
    command: str = "continue"


@dataclass
class ContinueResponseBody(Model):
    all_threads_continued: Optional[bool] = None


@dataclass
class _ContinueResponse(Model):
    body: ContinueResponseBody


@dataclass
class ContinueResponse(Response, _ContinueResponse):
    body: ContinueResponseBody = field()


@dataclass
class PauseArguments(Model):
    thread_id: int


@dataclass
class _PauseRequest(Model):
    arguments: PauseArguments


@dataclass
class PauseRequest(Request, _PauseRequest):
    arguments: PauseArguments = field()
    command: str = "pause"


@dataclass
class PauseResponse(Response):
    pass


@dataclass
class SteppingGranularity(Enum):
    STATEMENT = "statement"
    LINE = "line"
    INSTRUCTION = "instruction"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass
class NextArguments(Model):
    thread_id: int
    granularity: Optional[SteppingGranularity] = None


@dataclass
class _NextRequest(Model):
    arguments: NextArguments


@dataclass
class NextRequest(Request, _NextRequest):
    arguments: NextArguments = field()
    command: str = "next"


@dataclass
class NextResponse(Response):
    pass


@dataclass
class StepInArguments(Model):
    thread_id: int
    target_id: Optional[int] = None
    granularity: Optional[SteppingGranularity] = None


@dataclass
class _StepInRequest(Model):
    arguments: StepInArguments


@dataclass
class StepInRequest(Request, _StepInRequest):
    arguments: StepInArguments = field()
    command: str = "stepIn"


@dataclass
class StepInResponse(Response):
    pass


@dataclass
class StepOutArguments(Model):
    thread_id: int
    granularity: Optional[SteppingGranularity] = None


@dataclass
class _StepOutRequest(Model):
    arguments: StepOutArguments


@dataclass
class StepOutRequest(Request, _StepOutRequest):
    arguments: StepOutArguments = field()
    command: str = "stepOut"


@dataclass
class StepOutResponse(Response):
    pass


@dataclass
class ValueFormat(Model):
    hex: Optional[bool] = None


@dataclass
class VariablesArguments(Model):
    variables_reference: int
    filter: Optional[Literal["indexed", "named"]] = None
    start: Optional[int] = None
    count: Optional[int] = None
    format: Optional[ValueFormat] = None


@dataclass
class _VariablesRequest(Model):
    arguments: VariablesArguments


@dataclass
class VariablesRequest(Request, _VariablesRequest):
    arguments: VariablesArguments = field()
    command: str = "variables"


@dataclass
class VariablePresentationHint(Model):
    kind: Union[
        Literal[
            "property",
            "method",
            "class",
            "data",
            "event",
            "baseClass",
            "innerClass",
            "interface",
            "mostDerivedClass",
            "virtual",
            "dataBreakpoint",
        ],
        str,
        None,
    ] = None

    attributes: Optional[
        List[
            Union[
                Literal[
                    "static",
                    "constant",
                    "readOnly",
                    "rawString",
                    "hasObjectId",
                    "canHaveObjectId",
                    "hasSideEffects",
                    "hasDataBreakpoint",
                ],
                str,
            ]
        ]
    ] = None

    visibility: Union[Literal["public", "private", "protected", "internal", "final"], str, None] = None


@dataclass
class Variable(Model):
    name: str
    value: str
    type: Optional[str] = None
    presentation_hint: Optional[VariablePresentationHint] = None
    evaluate_name: Optional[str] = None
    variables_reference: int = 0
    named_variables: Optional[int] = None
    indexed_variables: Optional[int] = None
    memory_reference: Optional[str] = None


@dataclass
class VariablesResponseBody(Model):
    variables: List[Variable]


@dataclass
class _VariablesResponse(Model):
    body: VariablesResponseBody


@dataclass
class VariablesResponse(Response, _VariablesResponse):
    body: VariablesResponseBody = field()


class EvaluateArgumentContext(Enum):
    WATCH = "watch"
    REPL = "repl"
    HOVER = "hover"
    CLIPBOARD = "clipboard"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass
class EvaluateArguments(Model):
    expression: str
    frame_id: Optional[int] = None
    context: Union[EvaluateArgumentContext, str, None] = None
    format: Optional[ValueFormat] = None


@dataclass
class _EvaluateRequest(Model):
    arguments: EvaluateArguments


@dataclass
class EvaluateRequest(Request, _EvaluateRequest):
    arguments: EvaluateArguments = field()
    command: str = "evaluate"


@dataclass
class EvaluateResponseBody(Model):
    result: str
    type: Optional[str] = None
    presentation_hint: Optional[VariablePresentationHint] = None
    variables_reference: int = 0
    named_variables: Optional[int] = None
    indexed_variables: Optional[int] = None
    memory_reference: Optional[str] = None


@dataclass
class _EvaluateResponse(Model):
    body: VariablesResponseBody


@dataclass
class EvaluateResponse(Response, _EvaluateResponse):
    body: VariablesResponseBody = field()


@dataclass
class SetVariableArguments(Model):
    variables_reference: int
    name: str
    value: str
    format: Optional[ValueFormat] = None


@dataclass
class _SetVariableRequest(Model):
    arguments: SetVariableArguments


@dataclass
class SetVariableRequest(Request, _SetVariableRequest):
    arguments: SetVariableArguments = field()
    command: str = "setVariable"


@dataclass
class SetVariableResponseBody(Model):
    value: str
    type: Optional[str]
    variables_reference: Optional[int] = None
    named_variables: Optional[int] = None
    indexed_variables: Optional[int] = None


@dataclass
class _SetVariableResponse(Model):
    body: SetVariableResponseBody


@dataclass
class SetVariableResponse(Response, _SetVariableResponse):
    body: SetVariableResponseBody = field()


@dataclass(unsafe_hash=True)
class ExceptionFilterOptions(Model):
    filter_id: str
    condition: Optional[str] = None


class ExceptionBreakMode(Enum):
    NEVER = "never"
    ALWAYS = "always"
    UNHANDLED = "unhandled"
    USER_UNHANDLED = "userUnhandled"

    def __repr__(self) -> str:  # pragma: no cover
        return super().__str__()


@dataclass
class ExceptionPathSegment(Model):
    names: List[str]
    negate: Optional[bool] = None


@dataclass
class ExceptionOptions(Model):
    break_mode: ExceptionBreakMode
    path: Optional[List[ExceptionPathSegment]] = None


@dataclass
class SetExceptionBreakpointsArguments(Model):
    filters: List[str]
    filter_options: Optional[List[ExceptionFilterOptions]] = None
    exception_options: Optional[List[ExceptionOptions]] = None


@dataclass
class _SetExceptionBreakpointsRequest(Model):
    arguments: SetExceptionBreakpointsArguments


@dataclass
class SetExceptionBreakpointsRequest(Request, _SetExceptionBreakpointsRequest):
    arguments: SetExceptionBreakpointsArguments = field()
    command: str = "setExceptionBreakpoints"


@dataclass
class SetExceptionBreakpointsResponseBody(Model):
    breakpoints: Optional[List[Breakpoint]] = None


@dataclass
class SetExceptionBreakpointsResponse(Response):
    body: Optional[SetExceptionBreakpointsResponseBody] = None



from __future__ import annotations

import os
import re
from dataclasses import astuple, dataclass, fields
from pathlib import Path
from typing import Any, Iterator, Mapping, Optional, Union, overload
from urllib import parse

_IS_WIN = os.name == "nt"

_RE_DRIVE_LETTER_PATH = re.compile(r"^\/[a-zA-Z]:")

_DEFAULT_SCHEME = "file"


class InvalidUriError(Exception):
    pass


@dataclass
class _Parts:
    scheme: str = _DEFAULT_SCHEME
    netloc: str = ""
    path: str = ""
    params: str = ""
    query: str = ""
    fragment: str = ""

    def __iter__(self) -> Iterator[str]:
        yield from astuple(self)

    def __hash__(self) -> int:
        return hash((self.scheme, self.netloc, self.path, self.params, self.query, self.fragment))


class Uri(Mapping[str, str]):
    _parts: _Parts

    @overload
    def __init__(self, uri_str: str):
        ...

    @overload
    def __init__(self, *, scheme: str, netloc: str, path: str, params: str, query: str, fragment: str):
        ...

    def __init__(
        self,
        uri_str: Optional[str] = None,
        *,
        scheme: Optional[str] = None,
        netloc: Optional[str] = None,
        path: Optional[str] = None,
        params: Optional[str] = None,
        query: Optional[str] = None,
        fragment: Optional[str] = None,
    ) -> None:
        super().__init__()

        if uri_str is not None:
            self._parts = _Parts(*parse.urlparse(uri_str))
        else:
            self._parts = _Parts()

        if scheme is not None:
            self._parts.scheme = scheme
        if netloc is not None:
            self._parts.netloc = netloc
        if path is not None:
            self._parts.path = path
        if params is not None:
            self._parts.params = params
        if query is not None:
            self._parts.query = query
        if fragment is not None:
            self._parts.fragment = fragment

        self._parts.scheme = self._parts.scheme or _DEFAULT_SCHEME

    def __str__(self) -> str:
        return parse.urlunparse(tuple(self._parts))

    def __repr__(self) -> str:
        return f"{type(self).__name__}({repr(parse.urlunparse(tuple(self._parts)))})"

    def to_path(self) -> Path:
        return Path(self._to_path_str())

    def _to_path_str(self) -> str:
        """Returns the filesystem path of the given URI.
        Will handle UNC paths and normalize windows drive letters to lower-case. Also
        uses the platform specific path separator. Will *not* validate the path for
        invalid characters and semantics. Will *not* look at the scheme of this URI.
        """
        # scheme://netloc/path;parameters?query#fragment
        netloc = parse.unquote(self._parts.netloc)
        path = parse.unquote(self._parts.path)

        if self._parts.scheme != "file":
            raise InvalidUriError(f"Invalid URI scheme '{str(self)}'.")

        if netloc and self._parts.scheme == "file":
            # unc path: file://shares/c$/far/boo
            value = "//{}{}".format(netloc, path or "")

        elif _RE_DRIVE_LETTER_PATH.match(path):
            # windows drive letter: file:///C:/far/boo
            value = path[1].lower() + path[2:]

        else:
            # Other path
            value = path

        if _IS_WIN:
            value = value.replace("/", "\\")

        return value

    @property
    def scheme(self) -> str:
        return self._parts.scheme

    @property
    def netloc(self) -> str:
        return self._parts.netloc

    @property
    def path(self) -> str:
        return self._parts.path

    @property
    def params(self) -> str:
        return self._parts.params

    @property
    def query(self) -> str:
        return self._parts.query

    @staticmethod
    def from_path(path: Union[str, Path, os.PathLike[str]]) -> Uri:
        result = Uri(Path(path).as_uri())

        result._parts.path = parse.quote(parse.unquote(result._parts.path))
        return result

    def __eq__(self, o: object) -> bool:
        if isinstance(o, type(self)):
            return o._parts == self._parts

        return super().__eq__(o)

    def __hash__(self) -> int:
        return hash((type(self), self._parts))

    def __getitem__(self, k: str) -> Any:
        if any(e for e in fields(self._parts) if e.name == k):
            return getattr(self._parts, k)
        raise KeyError(k)

    def __len__(self) -> int:
        return len(fields(self._parts))

    def __iter__(self) -> Iterator[str]:
        for f in fields(self._parts):
            yield f.name

    def normalized(self) -> Uri:
        if self.scheme == "file":
            return Uri.from_path(self.to_path().resolve())

        return Uri(str(self))



from __future__ import annotations

import ast
import asyncio
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)

from ....utils.async_tools import threaded
from ....utils.logging import LoggingDescriptor
from ...common.decorators import language_id, retrigger_characters, trigger_characters
from ...common.lsp_types import (
    MarkupContent,
    MarkupKind,
    ParameterInformation,
    Position,
    SignatureHelp,
    SignatureHelpContext,
    SignatureInformation,
)
from ...common.text_document import TextDocument
from ..diagnostics.library_doc import KeywordDoc, LibraryDoc
from ..utils.ast import (
    Token,
    get_node_at_position,
    get_tokens_at_position,
    range_from_token,
    whitespace_at_begin_of_token,
    whitespace_from_begin_of_token,
)

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol

from .model_helper import ModelHelperMixin
from .protocol_part import RobotLanguageServerProtocolPart

_SignatureHelpMethod = Callable[
    [ast.AST, TextDocument, Position, Optional[SignatureHelpContext]], Awaitable[Optional[SignatureHelp]]
]


class RobotSignatureHelpProtocolPart(RobotLanguageServerProtocolPart, ModelHelperMixin):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        parent.signature_help.collect.add(self.collect)

    def _find_method(self, cls: Type[Any]) -> Optional[_SignatureHelpMethod]:
        if cls is ast.AST:
            return None
        method_name = "signature_help_" + cls.__name__
        if hasattr(self, method_name):
            method = getattr(self, method_name)
            if callable(method):
                return cast(_SignatureHelpMethod, method)
        for base in cls.__bases__:
            method = self._find_method(base)
            if method:
                return cast(_SignatureHelpMethod, method)
        return None

    @language_id("robotframework")
    @threaded()
    @trigger_characters([" ", "\t"])
    @retrigger_characters([" ", "\t"])
    @_logger.call
    async def collect(
        self, sender: Any, document: TextDocument, position: Position, context: Optional[SignatureHelpContext] = None
    ) -> Optional[SignatureHelp]:

        result_node = await get_node_at_position(await self.parent.documents_cache.get_model(document), position)
        if result_node is None:
            return None

        if result_node is None:
            return None

        method = self._find_method(type(result_node))
        if method is None:
            return None

        return await method(result_node, document, position, context)

    async def _signature_help_KeywordCall_or_Fixture(  # noqa: N802
        self,
        keyword_name_token_type: str,
        node: ast.AST,
        document: TextDocument,
        position: Position,
        context: Optional[SignatureHelpContext] = None,
    ) -> Optional[SignatureHelp]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Statement

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        kw_node = cast(Statement, node)

        tokens_at_position = get_tokens_at_position(kw_node, position)

        if not tokens_at_position:
            return None

        token_at_position = tokens_at_position[-1]

        if token_at_position.type not in [RobotToken.ARGUMENT, RobotToken.EOL, RobotToken.SEPARATOR]:
            return None

        token_at_position_index = kw_node.tokens.index(token_at_position)

        argument_token_index = token_at_position_index
        while argument_token_index >= 0 and kw_node.tokens[argument_token_index].type != RobotToken.ARGUMENT:
            argument_token_index -= 1

        arguments = kw_node.get_tokens(RobotToken.ARGUMENT)

        if argument_token_index >= 0:
            argument_token = kw_node.tokens[argument_token_index]
            if argument_token.type == RobotToken.ARGUMENT:
                argument_index = arguments.index(argument_token)
            else:
                argument_index = 0
        else:
            argument_index = -1

        if whitespace_at_begin_of_token(token_at_position) > 1:
            r = range_from_token(token_at_position)

            ws_b = whitespace_from_begin_of_token(token_at_position)
            r.start.character += 2 if ws_b and ws_b[0] != "\t" else 1

            if position.is_in_range(r) or r.end == position:
                argument_index += 1

        if argument_index < 0:
            return None

        result: Optional[Tuple[Optional[KeywordDoc], Token]] = None

        keyword_token = kw_node.get_token(keyword_name_token_type)
        if keyword_token is None:
            return None

        result = await self.get_keyworddoc_and_token_from_position(
            keyword_token.value,
            keyword_token,
            [cast(Token, t) for t in kw_node.get_tokens(RobotToken.ARGUMENT)],
            namespace,
            range_from_token(keyword_token).start,
            analyse_run_keywords=False,
        )

        if result is None or result[0] is None:
            return None

        if result[0].is_any_run_keyword():
            # TODO
            pass

        if (
            argument_index >= len(result[0].args)
            and len(result[0].args) > 0
            and not str(result[0].args[-1]).startswith("*")
        ):
            argument_index = -1

        signature = SignatureInformation(
            label=result[0].parameter_signature,
            parameters=[ParameterInformation(label=str(p)) for p in result[0].args],
            active_parameter=min(argument_index, len(result[0].args) - 1),
            documentation=MarkupContent(kind=MarkupKind.MARKDOWN, value=result[0].to_markdown(False)),
        )

        return SignatureHelp(
            signatures=[signature],
            active_signature=0,
            active_parameter=min(argument_index, len(result[0].args) - 1),
        )

    async def signature_help_KeywordCall(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position, context: Optional[SignatureHelpContext] = None
    ) -> Optional[SignatureHelp]:
        from robot.parsing.lexer.tokens import Token as RobotToken

        return await self._signature_help_KeywordCall_or_Fixture(RobotToken.KEYWORD, node, document, position, context)

    async def signature_help_Fixture(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position, context: Optional[SignatureHelpContext] = None
    ) -> Optional[SignatureHelp]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Fixture

        name_token = cast(Fixture, node).get_token(RobotToken.NAME)
        if name_token is None or name_token.value is None or name_token.value in ("", "NONE"):
            return None

        return await self._signature_help_KeywordCall_or_Fixture(RobotToken.NAME, node, document, position, context)

    async def signature_help_LibraryImport(  # noqa: N802
        self,
        node: ast.AST,
        document: TextDocument,
        position: Position,
        context: Optional[SignatureHelpContext] = None,
    ) -> Optional[SignatureHelp]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import LibraryImport

        # TODO from robot.utils.escaping import split_from_equals

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        library_node = cast(LibraryImport, node)

        if (
            library_node.name is None
            or position <= range_from_token(library_node.get_token(RobotToken.NAME)).extend(end_character=1).end
        ):
            return None

        lib_doc: Optional[LibraryDoc] = None
        try:
            lib_doc = await namespace.get_imported_library_libdoc(
                library_node.name, library_node.args, library_node.alias
            )

            if lib_doc is None or lib_doc.errors:
                lib_doc = await namespace.imports_manager.get_libdoc_for_library_import(
                    str(library_node.name), (), str(document.uri.to_path().parent)
                )

            if lib_doc is None:
                return None
        except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
            raise
        except BaseException:
            return None

        with_name_token = next((v for v in library_node.tokens if v.value == "WITH NAME"), None)
        if with_name_token is not None and position >= range_from_token(with_name_token).start:
            return None

        tokens_at_position = tokens_at_position = get_tokens_at_position(library_node, position)
        if not tokens_at_position:
            return None

        token_at_position = tokens_at_position[-1]

        if token_at_position.type not in [RobotToken.ARGUMENT, RobotToken.EOL, RobotToken.SEPARATOR]:
            return None

        token_at_position_index = library_node.tokens.index(token_at_position)

        argument_token_index = token_at_position_index
        while argument_token_index >= 0 and library_node.tokens[argument_token_index].type != RobotToken.ARGUMENT:
            argument_token_index -= 1

        arguments = library_node.get_tokens(RobotToken.ARGUMENT)

        if argument_token_index >= 0:
            argument_token = library_node.tokens[argument_token_index]
            if argument_token.type == RobotToken.ARGUMENT:
                argument_index = arguments.index(argument_token)
            else:
                argument_index = 0
        else:
            argument_index = -1

        if whitespace_at_begin_of_token(token_at_position) > 1:
            r = range_from_token(token_at_position)

            ws_b = whitespace_from_begin_of_token(token_at_position)
            r.start.character += 2 if ws_b and ws_b[0] != "\t" else 1

            if position.is_in_range(r) or r.end == position:
                argument_index += 1

        if argument_index < 0:
            return None

        signatures: List[SignatureInformation] = []

        # TODO check if we have a named argument
        # named_arg = False
        # arg_name: Optional[str] = None

        # if argument_index >= 0 and argument_index < len(arguments):
        #     name, value = split_from_equals(arguments[argument_index].value)
        #     if value is not None:
        #         arg_name = name
        #         named_arg = True

        for init in lib_doc.inits.values():
            if argument_index >= len(init.args) and len(init.args) > 0 and not str(init.args[-1]).startswith("*"):
                argument_index = -1

            signature = SignatureInformation(
                label=init.parameter_signature,
                parameters=[ParameterInformation(label=str(p)) for p in init.args],
                active_parameter=min(argument_index, len(init.args) - 1),
                documentation=MarkupContent(kind=MarkupKind.MARKDOWN, value=init.to_markdown(False)),
            )
            signatures.append(signature)

        if not signatures:
            return None

        return SignatureHelp(
            signatures=signatures,
            active_signature=0,
        )



from __future__ import annotations

from asyncio import CancelledError
from itertools import chain
from typing import TYPE_CHECKING, Any, List, Optional, Union, cast

from ....jsonrpc2.protocol import rpc_method
from ....utils.async_tools import async_tasking_event
from ....utils.logging import LoggingDescriptor
from ..decorators import (
    HasAllCommitCharacters,
    HasTriggerCharacters,
    language_id_filter,
)
from ..has_extend_capabilities import HasExtendCapabilities
from ..lsp_types import (
    CompletionContext,
    CompletionItem,
    CompletionList,
    CompletionOptions,
    CompletionParams,
    Position,
    ServerCapabilities,
    TextDocumentIdentifier,
)
from ..text_document import TextDocument

if TYPE_CHECKING:
    from ..protocol import LanguageServerProtocol

from .protocol_part import LanguageServerProtocolPart


class CompletionProtocolPart(LanguageServerProtocolPart, HasExtendCapabilities):

    _logger = LoggingDescriptor()

    def __init__(self, parent: LanguageServerProtocol) -> None:
        super().__init__(parent)

    @async_tasking_event
    async def collect(
        sender, document: TextDocument, position: Position, context: Optional[CompletionContext]  # NOSONAR
    ) -> Union[List[CompletionItem], CompletionList, None]:
        ...

    @async_tasking_event
    async def resolve(sender, completion_item: CompletionItem) -> CompletionItem:  # NOSONAR
        ...

    def extend_capabilities(self, capabilities: ServerCapabilities) -> None:
        if len(self.collect):
            trigger_chars = [
                k
                for k in chain(
                    *[
                        cast(HasTriggerCharacters, e).__trigger_characters__
                        for e in self.collect
                        if isinstance(e, HasTriggerCharacters)
                    ]
                )
            ]

            commit_chars = [
                k
                for k in chain(
                    *[
                        cast(HasAllCommitCharacters, e).__all_commit_characters__
                        for e in self.collect
                        if isinstance(e, HasAllCommitCharacters)
                    ]
                )
            ]
            capabilities.completion_provider = CompletionOptions(
                trigger_characters=trigger_chars if trigger_chars else None,
                all_commit_characters=commit_chars if commit_chars else None,
                resolve_provider=len(self.resolve) > 0,
                work_done_progress=True,
            )

    @rpc_method(name="textDocument/completion", param_type=CompletionParams)
    async def _text_document_completion(
        self,
        text_document: TextDocumentIdentifier,
        position: Position,
        context: Optional[CompletionContext],
        *args: Any,
        **kwargs: Any,
    ) -> Union[List[CompletionItem], CompletionList, None]:

        results: List[Union[List[CompletionItem], CompletionList]] = []

        document = await self.parent.documents.get(text_document.uri)
        if document is None:
            return None

        for result in await self.collect(
            self,
            document,
            position,
            context,
            callback_filter=language_id_filter(document),
        ):
            if isinstance(result, BaseException):
                if not isinstance(result, CancelledError):
                    self._logger.exception(result, exc_info=result)
            else:
                if result is not None:
                    results.append(result)

        if len(results) > 0:
            if any(e for e in results if isinstance(e, CompletionList)):
                result = CompletionList(
                    is_incomplete=any(e for e in results if isinstance(e, CompletionList) and e.is_incomplete),
                    items=[e for e in chain(*[r.items if isinstance(r, CompletionList) else r for r in results])],
                )
                if len(result.items) == 0:
                    return None
                return result
            else:
                result = [e for e in chain(*[k for k in results if isinstance(k, list)])]
                if len(result) == 0:
                    return None

                return result

        return None

    @rpc_method(name="completionItem/resolve", param_type=CompletionItem)
    async def _completion_item_resolve(
        self,
        params: CompletionItem,
        *args: Any,
        **kwargs: Any,
    ) -> CompletionItem:

        results: List[CompletionItem] = []

        for result in await self.resolve(self, params):
            if isinstance(result, BaseException):
                if not isinstance(result, CancelledError):
                    self._logger.exception(result, exc_info=result)
            else:
                if result is not None:
                    results.append(result)

        if len(results) > 0:
            if len(results) > 1:
                self._logger.warning("More then one resolve result. Use the last one.")

            return results[-1]

        return params



from typing import Any, Callable, List, Protocol, TypeVar, runtime_checkable

from robotcode.language_server.common.text_document import TextDocument

_F = TypeVar("_F", bound=Callable[..., Any])


def language_id(id: str) -> Callable[[_F], _F]:
    def decorator(func: _F) -> _F:
        setattr(func, "__language_id__", id)
        return func

    return decorator


@runtime_checkable
class HasLanguageId(Protocol):
    __language_id__: str


def trigger_characters(characters: List[str]) -> Callable[[_F], _F]:
    def decorator(func: _F) -> _F:
        setattr(func, "__trigger_characters__", characters)
        return func

    return decorator


@runtime_checkable
class HasRetriggerCharacters(Protocol):
    __retrigger_characters__: str


def retrigger_characters(characters: List[str]) -> Callable[[_F], _F]:
    def decorator(func: _F) -> _F:
        setattr(func, "__retrigger_characters__", characters)
        return func

    return decorator


@runtime_checkable
class HasTriggerCharacters(Protocol):
    __trigger_characters__: List[str]


def all_commit_characters(characters: List[str]) -> Callable[[_F], _F]:
    def decorator(func: _F) -> _F:
        setattr(func, "__all_commit_characters__", characters)
        return func

    return decorator


@runtime_checkable
class HasAllCommitCharacters(Protocol):
    __all_commit_characters__: List[str]


def language_id_filter(document: TextDocument) -> Callable[[Any], bool]:
    def filter(c: Any) -> bool:
        return not isinstance(c, HasLanguageId) or c.__language_id__ == document.language_id

    return filter



from __future__ import annotations

import ast
import asyncio
from typing import TYPE_CHECKING, Any, List, Optional

from ....utils.async_tools import check_canceled, threaded
from ....utils.logging import LoggingDescriptor
from ...common.decorators import language_id
from ...common.lsp_types import Diagnostic, DiagnosticSeverity, Position, Range
from ...common.parts.diagnostics import DiagnosticsResult
from ...common.text_document import TextDocument
from ..diagnostics.analyzer import Analyzer
from ..utils.ast import Token, range_from_node, range_from_token

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol

from .protocol_part import RobotLanguageServerProtocolPart


class RobotDiagnosticsProtocolPart(RobotLanguageServerProtocolPart):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        self.source_name = "robotcode.diagnostics"

        parent.diagnostics.collect.add(self.collect_token_errors)
        parent.diagnostics.collect.add(self.collect_walk_model_errors)

        parent.diagnostics.collect.add(self.collect_namespace_diagnostics)

        parent.documents_cache.namespace_invalidated.add(self.namespace_invalidated)

    @language_id("robotframework")
    async def namespace_invalidated(self, sender: Any, document: TextDocument) -> None:
        await self.parent.diagnostics.start_publish_diagnostics_task(document)

    @language_id("robotframework")
    @threaded()
    async def collect_namespace_diagnostics(self, sender: Any, document: TextDocument) -> DiagnosticsResult:
        try:
            namespace = await self.parent.documents_cache.get_namespace(document)
            if namespace is None:
                return DiagnosticsResult(self.collect_namespace_diagnostics, None)

            return DiagnosticsResult(self.collect_namespace_diagnostics, await namespace.get_diagnostisc())
        except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
            raise
        except BaseException as e:
            return DiagnosticsResult(
                self.collect_namespace_diagnostics,
                [
                    Diagnostic(
                        range=Range(
                            start=Position(
                                line=0,
                                character=0,
                            ),
                            end=Position(
                                line=len(await document.get_lines()),
                                character=len((await document.get_lines())[-1] or ""),
                            ),
                        ),
                        message=f"Fatal: can't get namespace diagnostics '{e}' ({type(e).__qualname__})",
                        severity=DiagnosticSeverity.ERROR,
                        source=self.source_name,
                        code=type(e).__qualname__,
                    )
                ],
            )

    def _create_error_from_node(self, node: ast.AST, msg: str, source: Optional[str] = None) -> Diagnostic:
        return Diagnostic(
            range=range_from_node(node, True),
            message=msg,
            severity=DiagnosticSeverity.ERROR,
            source=source if source is not None else self.source_name,
            code="ModelError",
        )

    def _create_error_from_token(self, token: Token, source: Optional[str] = None) -> Diagnostic:
        return Diagnostic(
            range=range_from_token(token),
            message=token.error if token.error is not None else "(No Message).",
            severity=DiagnosticSeverity.ERROR,
            source=source if source is not None else self.source_name,
            code="TokenError",
        )

    @language_id("robotframework")
    @_logger.call
    async def collect_token_errors(self, sender: Any, document: TextDocument) -> DiagnosticsResult:
        from robot.errors import VariableError
        from robot.parsing.lexer.tokens import Token

        result: List[Diagnostic] = []
        try:
            for token in await self.parent.documents_cache.get_tokens(document):
                await check_canceled()

                if token.type in [Token.ERROR, Token.FATAL_ERROR] and not await Analyzer.should_ignore(
                    document, range_from_token(token)
                ):
                    result.append(self._create_error_from_token(token))

                try:
                    for variable_token in token.tokenize_variables():
                        await check_canceled()
                        if variable_token == token:
                            break

                        if variable_token.type in [Token.ERROR, Token.FATAL_ERROR] and not await Analyzer.should_ignore(
                            document, range_from_token(variable_token)
                        ):
                            result.append(self._create_error_from_token(variable_token))

                except VariableError as e:
                    if not await Analyzer.should_ignore(document, range_from_token(token)):
                        result.append(
                            Diagnostic(
                                range=range_from_token(token),
                                message=str(e),
                                severity=DiagnosticSeverity.ERROR,
                                source=self.source_name,
                                code=type(e).__qualname__,
                            )
                        )
        except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
            raise
        except BaseException as e:
            return DiagnosticsResult(
                self.collect_token_errors,
                [
                    Diagnostic(
                        range=Range(
                            start=Position(
                                line=0,
                                character=0,
                            ),
                            end=Position(
                                line=len(await document.get_lines()),
                                character=len((await document.get_lines())[-1] or ""),
                            ),
                        ),
                        message=f"Fatal: can't get token diagnostics '{e}' ({type(e).__qualname__})",
                        severity=DiagnosticSeverity.ERROR,
                        source=self.source_name,
                        code=type(e).__qualname__,
                    )
                ],
            )

        return DiagnosticsResult(self.collect_token_errors, result)

    @language_id("robotframework")
    @threaded()
    @_logger.call
    async def collect_walk_model_errors(self, sender: Any, document: TextDocument) -> DiagnosticsResult:

        from ..utils.ast import HasError, HasErrors
        from ..utils.async_ast import iter_nodes

        try:
            model = await self.parent.documents_cache.get_model(document)

            result: List[Diagnostic] = []
            async for node in iter_nodes(model):
                error = node.error if isinstance(node, HasError) else None
                if error is not None and not await Analyzer.should_ignore(document, range_from_node(node)):
                    result.append(self._create_error_from_node(node, error))
                errors = node.errors if isinstance(node, HasErrors) else None
                if errors is not None:
                    for e in errors:
                        if not await Analyzer.should_ignore(document, range_from_node(node)):
                            result.append(self._create_error_from_node(node, e))

            return DiagnosticsResult(self.collect_walk_model_errors, result)

        except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
            raise
        except BaseException as e:
            return DiagnosticsResult(
                self.collect_walk_model_errors,
                [
                    Diagnostic(
                        range=Range(
                            start=Position(
                                line=0,
                                character=0,
                            ),
                            end=Position(
                                line=len(await document.get_lines()),
                                character=len((await document.get_lines())[-1] or ""),
                            ),
                        ),
                        message=f"Fatal: can't get model diagnostics '{e}' ({type(e).__qualname__})",
                        severity=DiagnosticSeverity.ERROR,
                        source=self.source_name,
                        code=type(e).__qualname__,
                    )
                ],
            )



from __future__ import annotations

from asyncio import CancelledError
from itertools import chain
from typing import TYPE_CHECKING, Any, List, Optional, cast

from ....jsonrpc2.protocol import rpc_method
from ....utils.async_tools import async_tasking_event
from ....utils.logging import LoggingDescriptor
from ..decorators import (
    HasRetriggerCharacters,
    HasTriggerCharacters,
    language_id_filter,
)
from ..has_extend_capabilities import HasExtendCapabilities
from ..lsp_types import (
    Position,
    ServerCapabilities,
    SignatureHelp,
    SignatureHelpContext,
    SignatureHelpOptions,
    SignatureHelpParams,
    TextDocumentIdentifier,
)
from ..text_document import TextDocument

if TYPE_CHECKING:
    from ..protocol import LanguageServerProtocol

from .protocol_part import LanguageServerProtocolPart


class SignatureHelpProtocolPart(LanguageServerProtocolPart, HasExtendCapabilities):

    _logger = LoggingDescriptor()

    def __init__(self, parent: LanguageServerProtocol) -> None:
        super().__init__(parent)

    @async_tasking_event
    async def collect(
        sender, document: TextDocument, position: Position, context: Optional[SignatureHelpContext] = None  # NOSONAR
    ) -> Optional[SignatureHelp]:
        ...

    def extend_capabilities(self, capabilities: ServerCapabilities) -> None:
        if len(self.collect):
            trigger_chars = [
                k
                for k in chain(
                    *[
                        cast(HasTriggerCharacters, e).__trigger_characters__
                        for e in self.collect
                        if isinstance(e, HasTriggerCharacters)
                    ]
                )
            ]

            retrigger_chars = [
                k
                for k in chain(
                    *[
                        cast(HasRetriggerCharacters, e).__retrigger_characters__
                        for e in self.collect
                        if isinstance(e, HasRetriggerCharacters)
                    ]
                )
            ]

            capabilities.signature_help_provider = SignatureHelpOptions(
                trigger_characters=trigger_chars if trigger_chars else None,
                retrigger_characters=retrigger_chars if retrigger_chars else None,
            )

    @rpc_method(name="textDocument/signatureHelp", param_type=SignatureHelpParams)
    async def _text_document_signature_help(
        self,
        text_document: TextDocumentIdentifier,
        position: Position,
        context: Optional[SignatureHelpContext] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Optional[SignatureHelp]:

        results: List[SignatureHelp] = []

        document = await self.parent.documents.get(text_document.uri)
        if document is None:
            return None

        for result in await self.collect(
            self,
            document,
            position,
            context,
            callback_filter=language_id_filter(document),
        ):
            if isinstance(result, BaseException):
                if not isinstance(result, CancelledError):
                    self._logger.exception(result, exc_info=result)
            else:
                if result is not None:
                    results.append(result)

        if len(results) > 0 and results[-1].signatures:
            # TODO: can we combine signature help results?

            return results[-1]

        return None
